//	   " ", "":

double RADIUS = 20;
double MASS = 8000;
double PX = 0, PY = 0, PZ = 0;
double VX = 0, VY = 0, VZ = 0;
int    MAXPOINTS = 200;
double DL = 10;
double ZOOM = 1;
AnsiString PlanetName = "";
bool ShowName = false;

//---------------------------------------------------------------------------
//          :
void LoadPlanet(int Sel)
{
	if (Sel<0 || Sel >= Universe->Count) return;
    MainForm->PanelColor->Color = Universe->Planets[Sel]->Color;
	MainForm->ColorDialog->Color = Universe->Planets[Sel]->Color;
	MainForm->ColorDialog->CustomColors->Clear();
	MainForm->ColorDialog->CustomColors->Add("ColorA=" + IntToHex(int(Universe->Planets[Sel]->Color), 6));
    MainForm->EditRadius->Text = Float2String(Universe->Planets[Sel]->Radius, RDigits);
    MainForm->EditMass->Text = Float2String(Universe->Planets[Sel]->Mass, RDigits);
    MainForm->EditX->Text = Float2String(Universe->Planets[Sel]->X, XDigits);
    MainForm->EditY->Text = Float2String(Universe->Planets[Sel]->Y, XDigits);
    MainForm->EditZ->Text = Float2String(Universe->Planets[Sel]->Z, XDigits);
    MainForm->EditVx->Text = Float2String(Universe->Planets[Sel]->Vx, XDigits);
    MainForm->EditVy->Text = Float2String(Universe->Planets[Sel]->Vy, XDigits);
    MainForm->EditVz->Text = Float2String(Universe->Planets[Sel]->Vz, XDigits);
    MainForm->SBtnTraj->Down = Universe->Planets[Sel]->HasTrajectory;
    RADIUS = Universe->Planets[Sel]->Radius;
    MASS = Universe->Planets[Sel]->Mass;
    PX = Universe->Planets[Sel]->X;
    PY = Universe->Planets[Sel]->Y;
    PZ = Universe->Planets[Sel]->Z;
    VX = Universe->Planets[Sel]->Vx;
    VY = Universe->Planets[Sel]->Vy;
    VZ = Universe->Planets[Sel]->Vz;
    MAXPOINTS = Universe->Planets[Sel]->Trajectory->GetMaxPoints();
    DL = Universe->Planets[Sel]->Trajectory->dL;
    PlanetName = Universe->Planets[Sel]->Name;
    ShowName = Universe->Planets[Sel]->ShowName;
    MainForm->LabelPlanet->Caption = (PlanetName != "") ? AnsiString(" ") + ":  " + PlanetName : AnsiString(" ");
    MainForm->ImagePlanet->Left = MainForm->LabelPlanet->Left + (MainForm->LabelPlanet->Width - MainForm->LabelPlanet->Canvas->TextWidth(MainForm->LabelPlanet->Caption))/2
                        - MainForm->ImagePlanet->Width - 5;
}
//---------------------------------------------------------------------------
//   :
int FindCenter(double *Xc, double *Yc, double *Zc)
{
	*Xc=0, *Yc=0, *Zc=0;		//    
    double M = 0, m;			//  
    for (int i=0; i<Universe->Count; i++)
		if (Universe->Planets[i]->Selected) {
        	m = Universe->Planets[i]->Mass;
            *Xc += (Universe->Planets[i]->X)*m;
            *Yc += (Universe->Planets[i]->Y)*m;
            *Zc += (Universe->Planets[i]->Z)*m;
            M += m;
        }
    if (M > MINDOUBLE) {
    	*Xc /= M;  *Yc /= M;  *Zc /= M;
        return 1;
    }
    return 0;
}
//---------------------------------------------------------------------------
//  ,     :
void TrackPlanet(double PX1,double PY1,double PZ1,double PX2,double PY2,double PZ2)
{
//	:
	if (Universe->Camera->RotateSight) {
		Universe->Camera->dX0 = 0;
		Universe->Camera->dY0 = 0;
		Universe->Camera->dZ0 = 0;
        double X1 = PX1 - Universe->Camera->X0;
        double Y1 = PY1 - Universe->Camera->Y0;
        double Z1 = PZ1 - Universe->Camera->Z0;
        double X2 = PX2 - Universe->Camera->X0;
        double Y2 = PY2 - Universe->Camera->Y0;
        double Z2 = PZ2 - Universe->Camera->Z0;
        //  :
        double Wx = Y1*Z2 - Y2*Z1;
        double Wy = X2*Z1 - X1*Z2;
        double Wz = X1*Y2 - X2*Y1;
        //  :
        double ct = (X1*X2 + Y1*Y2 + Z1*Z2)/sqrt((X1*X1 + Y1*Y1 + Z1*Z1)*(X2*X2 + Y2*Y2 + Z2*Z2));
        Universe->Camera->RotateAround(Wx,Wy,Wz,acos(ct));
    }
//	:
	double Xc=0, Yc=0, Zc=0;		//    
    if (FindCenter(&Xc, &Yc, &Zc)) {
        Universe->Camera->dX0 = Universe->Camera->X0 - Xc;
        Universe->Camera->dY0 = Universe->Camera->Y0 - Yc;
        Universe->Camera->dZ0 = Universe->Camera->Z0 - Zc;
    }
}
//---------------------------------------------------------------------------
//  ,      :
void TrackCenter(double Xc1, double Yc1, double Zc1)
{
	double Xc2=0, Yc2=0, Zc2=0;
    if (FindCenter(&Xc2, &Yc2, &Zc2)) {
	//	:
		if (Universe->Camera->RotateSight) {
			Universe->Camera->dX0 = 0;
			Universe->Camera->dY0 = 0;
			Universe->Camera->dZ0 = 0;
	        double X1 = Xc1 - Universe->Camera->X0;
	        double Y1 = Yc1 - Universe->Camera->Y0;
	        double Z1 = Zc1 - Universe->Camera->Z0;
	        double X2 = Xc2 - Universe->Camera->X0;
	        double Y2 = Yc2 - Universe->Camera->Y0;
	        double Z2 = Zc2 - Universe->Camera->Z0;
	        //  :
	        double Wx = Y1*Z2 - Y2*Z1;
	        double Wy = X2*Z1 - X1*Z2;
	        double Wz = X1*Y2 - X2*Y1;
	        //  :
    	    double ct = (X1*X2 + Y1*Y2 + Z1*Z2)/sqrt((X1*X1 + Y1*Y1 + Z1*Z1)*(X2*X2 + Y2*Y2 + Z2*Z2));
	        Universe->Camera->RotateAround(Wx,Wy,Wz,acos(ct));
    	}
    //	:
        Universe->Camera->dX0 = Universe->Camera->X0 - Xc2;
        Universe->Camera->dY0 = Universe->Camera->Y0 - Yc2;
        Universe->Camera->dZ0 = Universe->Camera->Z0 - Zc2;
    }
}
//---------------------------------------------------------------------------
//     dt:
void StrafeCamera(double dt)
{
    double **R = Universe->Camera->R;
    double Speed = Universe->Camera->Speed;
    double StrafeSpeed = Universe->Camera->StrafeSpeed;
    //	     :
    if (Universe->Camera->SpeedRelative) {
    	bool HasCenter = false;
    	for (int i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {HasCenter = true; break;}
        if (HasCenter) {
    		double L = sqrt((Universe->Camera->dX0)*(Universe->Camera->dX0) +
            	            (Universe->Camera->dY0)*(Universe->Camera->dY0) +
        	                (Universe->Camera->dZ0)*(Universe->Camera->dZ0) );
        	Speed = L;
        	StrafeSpeed = Speed/4;
        }
    }
    //	 :
    if (Numpad6) {
    	Universe->Camera->X0 += StrafeSpeed*R[0][0]*dt;
    	Universe->Camera->Y0 += StrafeSpeed*R[0][1]*dt;
    	Universe->Camera->Z0 += StrafeSpeed*R[0][2]*dt;
    	Universe->Camera->dX0 += StrafeSpeed*R[0][0]*dt;
    	Universe->Camera->dY0 += StrafeSpeed*R[0][1]*dt;
    	Universe->Camera->dZ0 += StrafeSpeed*R[0][2]*dt;
        NeedToDraw = true;
    }
    //	 :
    if (Numpad4) {
    	Universe->Camera->X0 -= StrafeSpeed*R[0][0]*dt;
    	Universe->Camera->Y0 -= StrafeSpeed*R[0][1]*dt;
    	Universe->Camera->Z0 -= StrafeSpeed*R[0][2]*dt;
    	Universe->Camera->dX0 -= StrafeSpeed*R[0][0]*dt;
    	Universe->Camera->dY0 -= StrafeSpeed*R[0][1]*dt;
    	Universe->Camera->dZ0 -= StrafeSpeed*R[0][2]*dt;
        NeedToDraw = true;
    }
    //	 :
    if (Numpad8) {
    	Universe->Camera->X0 += StrafeSpeed*R[1][0]*dt;
    	Universe->Camera->Y0 += StrafeSpeed*R[1][1]*dt;
    	Universe->Camera->Z0 += StrafeSpeed*R[1][2]*dt;
    	Universe->Camera->dX0 += StrafeSpeed*R[1][0]*dt;
    	Universe->Camera->dY0 += StrafeSpeed*R[1][1]*dt;
    	Universe->Camera->dZ0 += StrafeSpeed*R[1][2]*dt;
        NeedToDraw = true;
    }
    //	 :
    if (Numpad2) {
    	Universe->Camera->X0 -= StrafeSpeed*R[1][0]*dt;
    	Universe->Camera->Y0 -= StrafeSpeed*R[1][1]*dt;
    	Universe->Camera->Z0 -= StrafeSpeed*R[1][2]*dt;
    	Universe->Camera->dX0 -= StrafeSpeed*R[1][0]*dt;
    	Universe->Camera->dY0 -= StrafeSpeed*R[1][1]*dt;
    	Universe->Camera->dZ0 -= StrafeSpeed*R[1][2]*dt;
        NeedToDraw = true;
    }
    //  :
    if (Numpad5) {
    	Universe->Camera->X0 -= Speed*R[2][0]*dt;
    	Universe->Camera->Y0 -= Speed*R[2][1]*dt;
    	Universe->Camera->Z0 -= Speed*R[2][2]*dt;
    	Universe->Camera->dX0 -= Speed*R[2][0]*dt;
    	Universe->Camera->dY0 -= Speed*R[2][1]*dt;
    	Universe->Camera->dZ0 -= Speed*R[2][2]*dt;
        NeedToDraw = true;
    }
    //  :
    if (Numpad0) {
    	Universe->Camera->X0 += Speed*R[2][0]*dt;
    	Universe->Camera->Y0 += Speed*R[2][1]*dt;
    	Universe->Camera->Z0 += Speed*R[2][2]*dt;
    	Universe->Camera->dX0 += Speed*R[2][0]*dt;
    	Universe->Camera->dY0 += Speed*R[2][1]*dt;
    	Universe->Camera->dZ0 += Speed*R[2][2]*dt;
        NeedToDraw = true;
    }
}

//						  :

void __fastcall TMainForm::PanelColorMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	if (PLAY) return;
	if (ColorDialog->Execute()) {
    	PanelColor->Color = ColorDialog->Color;
        //	    :
        if (!SBtnGroup->Down) {
        	int Sel = Universe->Selected;
        	if (Sel > -1) {
        	    Universe->Planets[Sel]->Color = ColorDialog->Color;
                Universe->Planets[Sel]->TrajColor = ColorDialog->Color;
                Universe->Draw();
        	}
        }
        //	    :
        if (SBtnGroup->Down) {
        	bool Redraw = false;
    		for (int i=0; i<Universe->Count; i++)
        		if (Universe->Planets[i]->Selected) {
                	Universe->Planets[i]->Color = ColorDialog->Color;
	                Universe->Planets[i]->TrajColor = ColorDialog->Color;
                    Redraw = true;
                }
            if (Redraw) Universe->Draw();
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditRadiusEnter(TObject *Sender)
{
	EditRadius->Text = Float2String(RADIUS, MaxRDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditMassEnter(TObject *Sender)
{
	EditMass->Text = Float2String(MASS, MaxRDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditXEnter(TObject *Sender)
{
	EditX->Text = Float2String(PX, MaxXDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditYEnter(TObject *Sender)
{
	EditY->Text = Float2String(PY, MaxXDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditZEnter(TObject *Sender)
{
	EditZ->Text = Float2String(PZ, MaxXDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVxEnter(TObject *Sender)
{
	EditVx->Text = Float2String(VX, MaxXDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVyEnter(TObject *Sender)
{
	EditVy->Text = Float2String(VY, MaxXDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVzEnter(TObject *Sender)
{
	EditVz->Text = Float2String(VZ, MaxXDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditRadiusExit(TObject *Sender)
{
	if (!EditRadius->Modified) {
    	EditRadius->Text = Float2String(RADIUS, RDigits);
    	return;
    }
    else EditRadius->Modified = false;
	double Radius = 0;
    //	     :
	if (!sscanf(EditRadius->Text.c_str(), "%lf", &Radius) || (EditRadius->Text.Trim()).Length()==0 || Radius<=0) {
        EditRadius->Text = Float2String(RADIUS, RDigits);
        return;
    }
	//	:
    if (Radius == RADIUS) return;
    Radius = (Radius>MAXDOUBLE) ? MAXDOUBLE : Radius;
    EditRadius->Text = Float2String(Radius, RDigits);
    
	//	    :
	if (!SBtnGroup->Down && !PLAY) {
    	int Sel = Universe->Selected;
		if (Sel > -1) {
    	    Universe->Planets[Sel]->Radius = Radius;
    	    //	    ,    :
    	    if (!Universe->CheckPlanet(Universe->Planets[Sel], Sel) && Universe->Planets[Sel]->Radius > RADIUS) {
    	    	Universe->Planets[Sel]->Radius = RADIUS;
    	        EditRadius->Text = Float2String(RADIUS, RDigits);
                Radius = RADIUS;
    	    }
    		else Universe->Draw();
    	}
    }
    //	    :
    if (SBtnGroup->Down && !PLAY) {
    	bool Redraw = false;
    	if (!(!Universe->CheckAllSelected(1) && Radius>=RADIUS)) {
    		int i;
    		for (i=0; i<Universe->Count; i++)
        		if (Universe->Planets[i]->Selected) {
        	        Universe->Planets[i]->Radius = Radius;
                    Redraw = true;
                }
            if (!Redraw) goto RENEW;
        	//	     ,    :
        	if (!Universe->CheckAllSelected(1)) {
            	bool U = false;
            	double k = 0.9;
                double eps = 0.01;
        	    do {
                    if ((U && k<1) || (!U && k>1)) k = 1 + 0.5*(1-k);
        	    	Radius *= k;
    				for (i=0; i<Universe->Count; i++)
        				if (Universe->Planets[i]->Selected)
        	        		Universe->Planets[i]->Radius = Radius;
                    U = Universe->CheckAllSelected(1);
        	    } while (!(U && (-eps<1-k && 1-k<eps)) && Radius>MINDOUBLE);
			    EditRadius->Text = Float2String(Radius, RDigits);
        	}
        }
        else {
        	Radius = RADIUS;
            EditRadius->Text = Float2String(RADIUS, RDigits);
        }
        Universe->Draw();
    }
RENEW:
	//	:
    RADIUS = Radius;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditMassExit(TObject *Sender)
{
	if (!EditMass->Modified) {
    	EditMass->Text = Float2String(MASS, RDigits);
    	return;
    }
    else EditMass->Modified = false;
	double Mass = 0;
    //	     :
	if (!sscanf(EditMass->Text.c_str(), "%lf", &Mass) || (EditMass->Text.Trim()).Length()==0 || Mass<=0) {
        EditMass->Text = Float2String(MASS, RDigits);
        return;
    }
	//	:
    if (Mass == MASS) return;
    Mass = (Mass>MAXDOUBLE) ? MAXDOUBLE : Mass;
    MASS = Mass;
    EditMass->Text = Float2String(Mass, RDigits);
    if (PLAY) return;
    //	    :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1) {
        	Universe->Planets[Sel]->Mass = Mass;
        	TrackPlanet(PX,PY,PZ, PX,PY,PZ);
        }
    }
    //	    :
    if (SBtnGroup->Down) {
    	double X1=0, Y1=0, Z1=0;
        FindCenter(&X1, &Y1, &Z1);
    	for (int i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected)
                Universe->Planets[i]->Mass = Mass;
        TrackCenter(X1,Y1,Z1);
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditXExit(TObject *Sender)
{
	if (!EditX->Modified) {
    	EditX->Text = Float2String(PX, XDigits);
    	return;
    }
    else EditX->Modified = false;
	double x = 0;
    //	     :
	if (!sscanf(EditX->Text.c_str(), "%lf", &x) || (EditX->Text.Trim()).Length()==0) {
        EditX->Text = Float2String(PX, XDigits);
        return;
    }
    x = (x>MAXDOUBLE) ? MAXDOUBLE : x;
    x = (x<-MAXDOUBLE) ? -MAXDOUBLE : x;
    //	    :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1 && !PLAY) {
    	    Universe->Planets[Sel]->X = x;
    	    //	    ,  :
    	    if (!Universe->CheckPlanet(Universe->Planets[Sel], Sel)) {
    	    	Universe->Planets[Sel]->X = PX;
    	        x = PX;
    	    }
    	    Universe->Planets[Sel]->NextX = x;
            TrackPlanet(PX,PY,PZ, x,PY,PZ);
    	    Universe->Draw();
    	}
    }
    //	    :
    if (SBtnGroup->Down && !PLAY) {
    	double dx = x - PX;
        int i;
        double X1=0, Y1=0, Z1=0;
        FindCenter(&X1, &Y1, &Z1);
        bool Redraw = false;
    	for (i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
                Universe->Planets[i]->X += dx;
                Redraw = true;
            }
        if (!Redraw) goto RENEW;
        //      , :
        if (!Universe->CheckAllSelected(0)) {
    		for (i=0; i<Universe->Count; i++)
        		if (Universe->Planets[i]->Selected)
                	Universe->Planets[i]->X -= dx;
        	x = PX;  goto RENEW;
        }
    	for (i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
    	    	Universe->Planets[i]->NextX = Universe->Planets[i]->X;
            }
        TrackCenter(X1,Y1,Z1);
        Universe->Draw();
    }
RENEW:
	//	:
    PX = x;
    EditX->Text = Float2String(x, XDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditYExit(TObject *Sender)
{
	if (!EditY->Modified) {
    	EditY->Text = Float2String(PY, XDigits);
    	return;
    }
    else EditY->Modified = false;
	double y = 0;
    //	     :
	if (!sscanf(EditY->Text.c_str(), "%lf", &y) || (EditY->Text.Trim()).Length()==0) {
        EditY->Text = Float2String(PY, XDigits);
        return;
    }
    y = (y>MAXDOUBLE) ? MAXDOUBLE : y;
    y = (y<-MAXDOUBLE) ? -MAXDOUBLE : y;
    //	    :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1 && !PLAY) {
    	    Universe->Planets[Sel]->Y = y;
    	    //	    ,  :
    	    if (!Universe->CheckPlanet(Universe->Planets[Sel], Sel)) {
    	    	Universe->Planets[Sel]->Y = PY;
    	        y = PY;
    	    }
    	    Universe->Planets[Sel]->NextY = y;
            TrackPlanet(PX,PY,PZ, PX,y,PZ);
    	    Universe->Draw();
    	}
    }
    //	    :
    if (SBtnGroup->Down && !PLAY) {
    	double dy = y - PY;
        int i;
        double X1=0, Y1=0, Z1=0;
        FindCenter(&X1, &Y1, &Z1);
        bool Redraw = false;
    	for (i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
                Universe->Planets[i]->Y += dy;
                Redraw = true;
            }
        if (!Redraw) goto RENEW;
        //      , :
        if (!Universe->CheckAllSelected(0)) {
    		for (i=0; i<Universe->Count; i++)
        		if (Universe->Planets[i]->Selected)
                	Universe->Planets[i]->Y -= dy;
        	y = PY;  goto RENEW;
        }
    	for (i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
    	    	Universe->Planets[i]->NextY = Universe->Planets[i]->Y;
            }
        TrackCenter(X1,Y1,Z1);
        Universe->Draw();
    }
RENEW:
	//	:
    PY = y;
    EditY->Text = Float2String(y, XDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditZExit(TObject *Sender)
{
	if (!EditZ->Modified) {
    	EditZ->Text = Float2String(PZ, XDigits);
    	return;
    }
    else EditZ->Modified = false;
	double z = 0;
    //	     :
	if (!sscanf(EditZ->Text.c_str(), "%lf", &z) || (EditZ->Text.Trim()).Length()==0) {
        EditZ->Text = Float2String(PZ, XDigits);
        return;
    }
    z = (z>MAXDOUBLE) ? MAXDOUBLE : z;
    z = (z<-MAXDOUBLE) ? -MAXDOUBLE : z;
    //	    :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1 && !PLAY) {
    	    Universe->Planets[Sel]->Z = z;
    	    //	    ,  :
    	    if (!Universe->CheckPlanet(Universe->Planets[Sel], Sel)) {
    	    	Universe->Planets[Sel]->Z = PZ;
    	        z = PZ;
    	    }
    	    Universe->Planets[Sel]->NextZ = z;
            TrackPlanet(PX,PY,PZ, PX,PY,z);
    	    Universe->Draw();
    	}
    }
    //	    :
    if (SBtnGroup->Down && !PLAY) {
    	double dz = z - PZ;
        int i;
        double X1=0, Y1=0, Z1=0;
        FindCenter(&X1, &Y1, &Z1);
        bool Redraw = false;
    	for (i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
                Universe->Planets[i]->Z += dz;
                Redraw = true;
            }
        if (!Redraw) goto RENEW;
        //      , :
        if (!Universe->CheckAllSelected(0)) {
    		for (i=0; i<Universe->Count; i++)
        		if (Universe->Planets[i]->Selected)
                	Universe->Planets[i]->Z -= dz;
        	z = PZ;  goto RENEW;
        }
    	for (i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
    	    	Universe->Planets[i]->NextZ = Universe->Planets[i]->Z;
            }
        TrackCenter(X1,Y1,Z1);
        Universe->Draw();
    }
RENEW:
	//	:
    PZ = z;
    EditZ->Text = Float2String(z, XDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVxExit(TObject *Sender)
{
	if (!EditVx->Modified) {
    	EditVx->Text = Float2String(VX, XDigits);
    	return;
    }
    else EditVx->Modified = false;
	double Vx = 0;
    //	     :
	if (!sscanf(EditVx->Text.c_str(), "%lf", &Vx) || (EditVx->Text.Trim()).Length()==0) {
        EditVx->Text = Float2String(VX, XDigits);
        return;
    }
    Vx = (Vx>MAXDOUBLE) ? MAXDOUBLE : Vx;
    Vx = (Vx<-MAXDOUBLE) ? -MAXDOUBLE : Vx;
    //	   :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1 && !PLAY) {
    	    Universe->Planets[Sel]->Vx = Vx;
    	    Universe->Planets[Sel]->NextVx = Vx;
            Universe->Draw();
    	}
    }
    //	    :
    if (SBtnGroup->Down && !PLAY) {
    	double dVx = Vx - VX;
        bool Redraw = false;
    	for (int i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
    	    	Universe->Planets[i]->Vx += dVx;
    	    	Universe->Planets[i]->NextVx = Universe->Planets[i]->Vx;
    	    	Redraw = true;
            }
        if (Redraw) Universe->Draw();
    }
	//	:
    VX = Vx;
    EditVx->Text = Float2String(Vx, XDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVyExit(TObject *Sender)
{
	if (!EditVy->Modified) {
    	EditVy->Text = Float2String(VY, XDigits);
    	return;
    }
    else EditVy->Modified = false;
	double Vy = 0;
    //	     :
	if (!sscanf(EditVy->Text.c_str(), "%lf", &Vy) || (EditVy->Text.Trim()).Length()==0) {
        EditVy->Text = Float2String(VY, XDigits);
        return;
    }
    Vy = (Vy>MAXDOUBLE) ? MAXDOUBLE : Vy;
    Vy = (Vy<-MAXDOUBLE) ? -MAXDOUBLE : Vy;
    //	   :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1 && !PLAY) {
    	    Universe->Planets[Sel]->Vy = Vy;
    	    Universe->Planets[Sel]->NextVy = Vy;
            Universe->Draw();
    	}
    }
    //	    :
    if (SBtnGroup->Down && !PLAY) {
    	double dVy = Vy - VY;
        bool Redraw = false;
    	for (int i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
    	    	Universe->Planets[i]->Vy += dVy;
    	    	Universe->Planets[i]->NextVy = Universe->Planets[i]->Vy;
    	    	Redraw = true;
            }
        if (Redraw) Universe->Draw();
    }
	//	:
    VY = Vy;
    EditVy->Text = Float2String(Vy, XDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVzExit(TObject *Sender)
{
	if (!EditVz->Modified) {
    	EditVz->Text = Float2String(VZ, XDigits);
    	return;
    }
    else EditVz->Modified = false;
	double Vz = 0;
    //	     :
	if (!sscanf(EditVz->Text.c_str(), "%lf", &Vz) || (EditVz->Text.Trim()).Length()==0) {
        EditVz->Text = Float2String(VZ, XDigits);
        return;
    }
    Vz = (Vz>MAXDOUBLE) ? MAXDOUBLE : Vz;
    Vz = (Vz<-MAXDOUBLE) ? -MAXDOUBLE : Vz;
    //	   :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1 && !PLAY) {
    	    Universe->Planets[Sel]->Vz = Vz;
    	    Universe->Planets[Sel]->NextVz = Vz;
            Universe->Draw();
    	}
    }
    //	    :
    if (SBtnGroup->Down && !PLAY) {
    	double dVz = Vz - VZ;
        bool Redraw = false;
    	for (int i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
    	    	Universe->Planets[i]->Vz += dVz;
    	    	Universe->Planets[i]->NextVz = Universe->Planets[i]->Vz;
    	    	Redraw = true;
            }
        if (Redraw) Universe->Draw();
    }
	//	:
    VZ = Vz;
    EditVz->Text = Float2String(Vz, XDigits);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LabelPlanetClick(TObject *Sender)
{
	if (PLAY) return;
    if (EditZoom->Focused()) { EditZoomExit(Sender); EditZoom->SelectAll(); }
    Application->CreateForm(__classid(TPlanetBox), &PlanetBox);
    TPoint XYPanel;  XYPanel.x = 0;  XYPanel.y = PanelPlanet->Height;
    TPoint XYScreen = PanelPlanet->ClientToScreen(XYPanel);
    PlanetBox->Left = XYScreen.x-1;  PlanetBox->Top = XYScreen.y;
    PlanetBox->Width = PanelPlanet->Width + 1;
	PlanetBox->ShowModal();

    //	   :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
        if (Sel > -1) {
        	Universe->Planets[Sel]->Name = PlanetName;
            Universe->Planets[Sel]->ShowName = ShowName;
            Universe->Draw();
        }
    }
    //	    :
    if (SBtnGroup->Down) {
        bool Redraw = false;
    	for (int i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected) {
            	Universe->Planets[i]->ShowName = ShowName;
            	Redraw = true;
            }
        if (Redraw) Universe->Draw();
    }
    LabelPlanet->Caption = (PlanetName != "") ? AnsiString(" ") + ":  " + PlanetName : AnsiString(" ");
    ImagePlanet->Left = LabelPlanet->Left + (LabelPlanet->Width - LabelPlanet->Canvas->TextWidth(LabelPlanet->Caption))/2
                        - ImagePlanet->Width - 5;
}
//---------------------------------------------------------------------------

//                          :

void __fastcall TMainForm::EditZoomExit(TObject *Sender)
{
	if (!EditZoom->Modified) return;
    else EditZoom->Modified = false;
	double Zoom = 0;
    //	     :
	if (!sscanf(EditZoom->Text.c_str(), "%lf", &Zoom) || (EditZoom->Text.Trim()).Length()==0) {
        EditZoom->Text = Float2String(ZOOM,8);
        return;
    }
    if (Zoom < 1) {
        EditZoom->Text = Float2String(ZOOM,8);
        return;
    }
    Zoom = (Zoom>MAXDOUBLE) ? MAXDOUBLE : Zoom;
	//	  :
    double F0 = double(Image->Width)/2/tan(Universe->Camera->MaxPhi);
    Universe->Camera->Phi = atan(Image->Width/2/(Zoom*F0));
    if (!PLAY) Universe->Draw();
    ZOOM = Zoom;
    EditZoom->Text = Float2String(Zoom, 8);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SBtnZoomAllClick(TObject *Sender)
{
	if (Universe->Count == 0) return;
    double X0 = Universe->Camera->X0;
    double Y0 = Universe->Camera->Y0;
    double Z0 = Universe->Camera->Z0;
    double **R = Universe->Camera->R;
    double F = double(Image->Width)/2/tan(Universe->Camera->Phi);
    double XR,YR,ZR;
    int i;

//	       :
	double MaxZR = -MAXDOUBLE;
    for (i=0; i<Universe->Count; i++) {
    	ZR = R[2][0]*(Universe->Planets[i]->X-X0) + R[2][1]*(Universe->Planets[i]->Y-Y0) + R[2][2]*(Universe->Planets[i]->Z-Z0);
    	if (ZR >= MaxZR) MaxZR = ZR;
    }
//	    MaxZR + F:
	if (MaxZR > -F) {
	   	Universe->Camera->X0 += (MaxZR + F)*R[2][0];
	   	Universe->Camera->Y0 += (MaxZR + F)*R[2][1];
	   	Universe->Camera->Z0 += (MaxZR + F)*R[2][2];
	   	Universe->Camera->dX0 += (MaxZR + F)*R[2][0];
	   	Universe->Camera->dY0 += (MaxZR + F)*R[2][1];
	   	Universe->Camera->dZ0 += (MaxZR + F)*R[2][2];
	    X0 = Universe->Camera->X0;
	    Y0 = Universe->Camera->Y0;
	    Z0 = Universe->Camera->Z0;
    }
//	     :
	double MaxXR=0, MaxYR=0;
    double ZRMaxXR,ZRMaxYR;
    for (i=0; i<Universe->Count; i++) {
    	XR = R[0][0]*(Universe->Planets[i]->X-X0) + R[0][1]*(Universe->Planets[i]->Y-Y0) + R[0][2]*(Universe->Planets[i]->Z-Z0);
    	YR = R[1][0]*(Universe->Planets[i]->X-X0) + R[1][1]*(Universe->Planets[i]->Y-Y0) + R[1][2]*(Universe->Planets[i]->Z-Z0);
    	ZR = R[2][0]*(Universe->Planets[i]->X-X0) + R[2][1]*(Universe->Planets[i]->Y-Y0) + R[2][2]*(Universe->Planets[i]->Z-Z0);
        XR = (XR>0) ? XR : -XR;
        YR = (YR>0) ? YR : -YR;
        if (XR >= MaxXR) { MaxXR = XR;  ZRMaxXR = -ZR; }
        if (YR >= MaxYR) { MaxYR = YR;  ZRMaxYR = -ZR; }
    }
//	    :
	double tanPhiW = double(Image->Width)/2/F;
	double tanPhiH = double(Image->Height)/2/F;
    double dZRW = MaxXR/(0.9*tanPhiW) - ZRMaxXR;
    double dZRH = MaxYR/(0.9*tanPhiH) - ZRMaxYR;
    double dZR = max(dZRW,dZRH);
//	  :
	if (dZR > 0) {
	   	Universe->Camera->X0 += dZR*R[2][0];
	   	Universe->Camera->Y0 += dZR*R[2][1];
	   	Universe->Camera->Z0 += dZR*R[2][2];
	   	Universe->Camera->dX0 += dZR*R[2][0];
	   	Universe->Camera->dY0 += dZR*R[2][1];
	   	Universe->Camera->dZ0 += dZR*R[2][2];
    }
    if (!PLAY) Universe->Draw();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SBtnZoomCenterClick(TObject *Sender)
{
//	    :
	double Xc=0, Yc=0, Zc=0;
    if (!FindCenter(&Xc,&Yc,&Zc)) return;
//	 . :
	double MaxR = 0;
	for (int i=0; i<Universe->Count; i++)
    	if (Universe->Planets[i]->Selected &&
            Universe->Planets[i]->Radius > MaxR) MaxR = Universe->Planets[i]->Radius;
//	     . :
	int nR = 10;
    double L = sqrt( (Universe->Camera->dX0)*(Universe->Camera->dX0) +
                     (Universe->Camera->dY0)*(Universe->Camera->dY0) +
                     (Universe->Camera->dZ0)*(Universe->Camera->dZ0) );
    Universe->Camera->dX0 *= nR*MaxR/L;
    Universe->Camera->dY0 *= nR*MaxR/L;
    Universe->Camera->dZ0 *= nR*MaxR/L;
    Universe->Camera->X0 = Xc + Universe->Camera->dX0;
    Universe->Camera->Y0 = Yc + Universe->Camera->dY0;
    Universe->Camera->Z0 = Zc + Universe->Camera->dZ0;
    if (!PLAY) Universe->Draw();
}
//---------------------------------------------------------------------------

//					  :

void __fastcall TMainForm::EditRadiusKeyDown(TObject *Sender,
      WORD &Key, TShiftState Shift)
{
	if (Key == VK_RETURN) EditMass->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditMassKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditX->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditXKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditY->SetFocus();	
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditYKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditZ->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditZKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditVx->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVxKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditVy->SetFocus();	
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVyKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditVz->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditVzKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditRadius->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditZoomKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) { EditZoomExit(Sender); EditZoom->SelectAll(); }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EditMassDblClick(TObject *Sender)
{
	MASS = RADIUS*RADIUS*RADIUS;
    EditMass->Text = Float2String(MASS, RDigits);
    if (PLAY) return;
    //	    :
    if (!SBtnGroup->Down) {
    	int Sel = Universe->Selected;
    	if (Sel>-1) Universe->Planets[Sel]->Mass = MASS;
    }
    //	    :
    if (SBtnGroup->Down) {
    	for (int i=0; i<Universe->Count; i++)
        	if (Universe->Planets[i]->Selected)
                Universe->Planets[i]->Mass = MASS;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SBtnTrajClick(TObject *Sender)
{
    //	 :
	if (SBtnTraj->Down) {
    	//	  :
    	if (!SBtnGroup->Down) {
    		int Sel = Universe->Selected;
        	if (Sel > -1) Universe->Planets[Sel]->HasTrajectory = true;
        }
        //	   :
    	if (SBtnGroup->Down) {
        	for (int i=0; i<Universe->Count; i++)
        		if (Universe->Planets[i]->Selected)
                	Universe->Planets[i]->HasTrajectory = true;
        }
    }

    //	 :
    if (!SBtnTraj->Down) {
    	//	  :
    	if (!SBtnGroup->Down) {
    		int Sel = Universe->Selected;
        	if (Sel > -1) {
            	Universe->Planets[Sel]->HasTrajectory = false;
                Universe->Planets[Sel]->Trajectory->Init();
            }
        }
        //	   :
    	if (SBtnGroup->Down) {
        	for (int i=0; i<Universe->Count; i++)
        		if (Universe->Planets[i]->Selected) {
                	Universe->Planets[i]->HasTrajectory = false;
                	Universe->Planets[i]->Trajectory->Init();
                }
        }
    }

    if (!PLAY) Universe->Draw();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	Shft = Shift.Contains(ssShift);
    Ctrl = Shift.Contains(ssCtrl);
    Numpad0 = (Key == VK_NUMPAD0) || Numpad0;
    Numpad2 = (Key == VK_NUMPAD2) || Numpad2;
    Numpad4 = (Key == VK_NUMPAD4) || Numpad4;
    Numpad5 = (Key == VK_NUMPAD5) || Numpad5;
    Numpad6 = (Key == VK_NUMPAD6) || Numpad6;
    Numpad8 = (Key == VK_NUMPAD8) || Numpad8;

    if (SelectingRect) return;

    //  :
    if ((!RightMouse && Ctrl && !Shft) || (!LeftMouse && !RightMouse)) {
    	int Touched = Universe->Touched(MX,MY);					//  
        if (Touched > -1) Screen->Cursor = crHandPoint;
        else Screen->Cursor = crDefault;
    }
    else if (!RightMouse) Screen->Cursor = crDefault;
    if (LeftMouse &&  Ctrl &&  Shft) Screen->Cursor = crSizeWE;	// 
    if (LeftMouse && !Ctrl) Screen->Cursor = crHandPoint;		// 
    if (RightMouse) Screen->Cursor = crSizeNS;					//   

    //  :
	if (Panel->Focused()) {
    	if (Key == VK_DELETE && !PLAY)
        	SBtnDelete->Click();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormKeyUp(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	Shft = Shift.Contains(ssShift);
    Ctrl = Shift.Contains(ssCtrl);
    Numpad0 = !(Numpad0 && (Key == VK_NUMPAD0)) && Numpad0;
    Numpad2 = !(Numpad2 && (Key == VK_NUMPAD2)) && Numpad2;
    Numpad4 = !(Numpad4 && (Key == VK_NUMPAD4)) && Numpad4;
    Numpad5 = !(Numpad5 && (Key == VK_NUMPAD5)) && Numpad5;
    Numpad6 = !(Numpad6 && (Key == VK_NUMPAD6)) && Numpad6;
    Numpad8 = !(Numpad8 && (Key == VK_NUMPAD8)) && Numpad8;

    if (SelectingRect) return;

    //  :
    if ((!RightMouse && Ctrl && !Shft) || (!LeftMouse && !RightMouse)) {
    	int Touched = Universe->Touched(MX,MY);					//  
        if (Touched > -1) Screen->Cursor = crHandPoint;
        else Screen->Cursor = crDefault;
    }
    else if (!RightMouse) Screen->Cursor = crDefault;
    if (LeftMouse &&  Ctrl &&  Shft) Screen->Cursor = crSizeWE;	// 
    if (LeftMouse && !Ctrl) Screen->Cursor = crHandPoint;		// 
    if (RightMouse) Screen->Cursor = crSizeNS;					//   
}
//---------------------------------------------------------------------------

