//---------------------------------------------------------------------------
#include <vcl.h>
#include <stdio.h>
#include "TUniverse.cpp"
#pragma hdrstop

#include "TCameraBox.h"
#include "Unit1.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TCameraBox *CameraBox;
extern TUniverse *Universe;

double FBSPEED, SPEED, MAXPHI, CX0, CY0, CZ0;
double CDigits = 10;

//---------------------------------------------------------------------------
void EditExit(TEdit *Edit, double *VAR, bool MoreThanZero)
{
	if (!Edit->Modified) return;
    else Edit->Modified = false;
	double Var = 0;
    //	     :
	if (!sscanf(Edit->Text.c_str(), "%lf", &Var) || (Edit->Text.Trim()).Length()==0) {
        Edit->Text = Float2String(*VAR,CDigits);
        return;
    }
    if (MoreThanZero && Var <= 0) {
        Edit->Text = Float2String(*VAR,CDigits);
        return;
    }
	//	  :
    *VAR = Var;
    Edit->Text = Float2String(Var,CDigits);
}
//---------------------------------------------------------------------------
__fastcall TCameraBox::TCameraBox(TComponent* Owner)
	: TForm(Owner)
{
    FBSPEED = Universe->Camera->Speed;
	SPEED = Universe->Camera->StrafeSpeed;
    MAXPHI = 2*(Universe->Camera->MaxPhi)/M_PI*180;
    CX0 = Universe->Camera->X0;
    CY0 = Universe->Camera->Y0;
    CZ0 = Universe->Camera->Z0;
    EditFBSpeed->Text = Float2String(FBSPEED,CDigits);
    EditStrafeSpeed->Text = Float2String(SPEED,CDigits);
    EditMaxPhi->Text = Float2String(MAXPHI, CDigits);
    EditX0->Text = Float2String(CX0,CDigits);
    EditY0->Text = Float2String(CY0,CDigits);
    EditZ0->Text = Float2String(CZ0,CDigits);
    CheckBoxRotateSight->Checked = Universe->Camera->RotateSight;
    CheckBoxRotateWith->Checked = Universe->Camera->RotateWith;
    CheckBoxDrawAxes->Checked = Universe->Camera->DrawAxes;
    CheckBoxDrawPoint->Checked = Universe->Camera->DrawPoint;
    CheckBoxSpeedRelative->Checked = Universe->Camera->SpeedRelative;
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	Release();	
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::BtnCancelClick(TObject *Sender)
{
	Close();	
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditFBSpeedExit(TObject *Sender)
{
	EditExit(EditFBSpeed, &FBSPEED, true);
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditStrafeSpeedExit(TObject *Sender)
{
	EditExit(EditStrafeSpeed, &SPEED, true);
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditMaxPhiExit(TObject *Sender)
{
	EditExit(EditMaxPhi, &MAXPHI, true);
    if (MAXPHI >= 180) {
    	MAXPHI = 179;
    	EditMaxPhi->Text = Float2String(MAXPHI, CDigits);
    }
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditX0Exit(TObject *Sender)
{
	EditExit(EditX0, &CX0, false);
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditY0Exit(TObject *Sender)
{
	EditExit(EditY0, &CY0, false);
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditZ0Exit(TObject *Sender)
{
	EditExit(EditZ0, &CZ0, false);
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditFBSpeedKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditStrafeSpeed->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditStrafeSpeedKeyDown(TObject *Sender,
      WORD &Key, TShiftState Shift)
{
	if (Key == VK_RETURN) EditMaxPhi->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditMaxPhiKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditX0->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditX0KeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditY0->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditY0KeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) EditZ0->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::EditZ0KeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
	if (Key == VK_RETURN) CheckBoxXY->SetFocus();
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::CheckBoxXYClick(TObject *Sender)
{
	CheckBoxXZ->Checked = false;
	CheckBoxYZ->Checked = false;
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::CheckBoxXZClick(TObject *Sender)
{
	CheckBoxXY->Checked = false;
	CheckBoxYZ->Checked = false;
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::CheckBoxYZClick(TObject *Sender)
{
	CheckBoxXY->Checked = false;
	CheckBoxXZ->Checked = false;
}
//---------------------------------------------------------------------------
void __fastcall TCameraBox::BtnOKClick(TObject *Sender)
{
	Universe->Camera->dX0 = CX0 - (Universe->Camera->X0 - Universe->Camera->dX0);
	Universe->Camera->dY0 = CY0 - (Universe->Camera->Y0 - Universe->Camera->dY0);
	Universe->Camera->dZ0 = CZ0 - (Universe->Camera->Z0 - Universe->Camera->dZ0);
    Universe->Camera->X0 = CX0;
    Universe->Camera->Y0 = CY0;
    Universe->Camera->Z0 = CZ0;

    Universe->Camera->Speed = FBSPEED;
    Universe->Camera->StrafeSpeed = SPEED;
    double F0 = double(MainForm->Image->Width)/2/tan(Universe->Camera->MaxPhi);
    double F = double(MainForm->Image->Width)/2/tan(Universe->Camera->Phi);
    double Zoom = F/F0;
    Universe->Camera->MaxPhi = MAXPHI/2/180*M_PI;
    F0 = double(MainForm->Image->Width)/2/tan(Universe->Camera->MaxPhi);
    F = Zoom*F0;
    Universe->Camera->Phi = atan(double(MainForm->Image->Width)/2/F);

    Universe->Camera->RotateSight = CheckBoxRotateSight->Checked;
    Universe->Camera->RotateWith = CheckBoxRotateWith->Checked;
    Universe->Camera->DrawAxes = CheckBoxDrawAxes->Checked;
    Universe->Camera->DrawPoint = CheckBoxDrawPoint->Checked;
    Universe->Camera->SpeedRelative = CheckBoxSpeedRelative->Checked;

    double **R = Universe->Camera->R;
    if (CheckBoxXY->Checked) {
    	R[0][0]=1;  R[0][1]=0;  R[0][2]=0;
    	R[1][0]=0;  R[1][1]=1;  R[1][2]=0;
    	R[2][0]=0;  R[2][1]=0;  R[2][2]=1;
    }
    if (CheckBoxXZ->Checked) {
    	R[0][0]=1;  R[0][1]=0;   R[0][2]=0;
    	R[1][0]=0;  R[1][1]=0;   R[1][2]=1;
    	R[2][0]=0;  R[2][1]=-1;  R[2][2]=0;
    }
    if (CheckBoxYZ->Checked) {
    	R[0][0]=0;  R[0][1]=1;  R[0][2]=0;
    	R[1][0]=0;  R[1][1]=0;  R[1][2]=1;
    	R[2][0]=1;  R[2][1]=0;  R[2][2]=0;
    }
    Close();
}
//---------------------------------------------------------------------------
