#include <alloc.h>
#include "TPlanet.cpp"
#include "TTimeEx.cpp"
#include "TreeSort.cpp"

// ---------------------------   --------------------------------

class TUniverse {
private:

	int		Pages;				//    
    int		PageSize;			//    ( )
    int     *Numbers;			//   , .  
    double  *Criteria;			//   , .  
    bool	Draw3D;				//   : false - , true - 3D-
    Byte	**BMP;              //     

    int		NRegs;				//      - ()
								// (0-    + .   )
    double	**dx,**dy,**dz;		//      
    double	**dVx,**dVy,**dVz;	//      
    double	*c4,*c5,*c6,*c7,*c8;//     :
    double	*b4,*b5,*b6,*b7,*b8;
    double	**a4,**a5,**a6,**a7,**a8;

    double  Tau;				//      Next_
    bool	**NoImpacts;		// ;  [i][j]-  true,   
    							//  i-  j-   

    void RungeKutta2Step(double);	//      2(2)
    void RungeKutta3Step(double);	//      3(3)
    void RungeKutta4Step(double);	//      -
    void XRungeKuttaStep(double*, double*, double**, int, double);
    								//      -
    void Impact(int,int);			//    
    int  MoveTo1stImpact();			//     

public:

	TPlanet **Planets;			//  
    int		Count;				//  
    int		Active;				//   
    int		Selected;			//    
    bool	Design;				//   design-time
    Byte	Brightness;			//  3D-

    TImage	*Image;				//     
    CAMERA	*Camera;			// ,     
    TTimeEx *Time;				//   

	double	Gamma;				//  
    int     DetectImpacts;		//   
    double	Alpha;				//  ,  0 = .  ,  1 = . 
    int		Method;				//  
    TColor	Color;				//  

    TUniverse();
	~TUniverse();
    void SetDraw3D(bool);			//   3D-
    bool GetDraw3D();				//   3D-
    int  NewPlanet(TColor,double,double,double,double,double,double,double,double,bool);
    								//     
    int  NewPlanet(double,bool,double,double);
									//   
    void RandomPlanets(int,bool,double,double,double,double,double,double,double,double);
    								//      
    void RandomPlanets(int,double,double,double,double,double,double,double,double,
                       double,double,double,double,double,double);
    								//      
    void DeletePlanet(int);			//  
    void SelectPlanet(int);			//  
    void DeselectAll();				//     
    void ActivatePlanet(int);		//  
    int  CheckPlanet(TPlanet*,int);	//     
    int  CheckAllSelected(int);		//    
    void UpdateSelectRect(TRect*);	//   
    void Init();					//   
    void Draw();					// 
    int  Touched(int,int);			//   
    void NextStep(double);			//  
};

// ----------------------------------------------------------------------------
// :

TUniverse::TUniverse()
{
	PageSize = 64;
    Planets = (TPlanet **) calloc(PageSize,sizeof(TPlanet *));
    Numbers =  (int *)    calloc(PageSize,sizeof(int));
    Criteria = (double *) calloc(PageSize,sizeof(double));

    NRegs = 13;	//     = 12
    dx  = (double **) calloc(PageSize,sizeof(double *));
    dy  = (double **) calloc(PageSize,sizeof(double *));
    dz  = (double **) calloc(PageSize,sizeof(double *));
    dVx = (double **) calloc(PageSize,sizeof(double *));
    dVy = (double **) calloc(PageSize,sizeof(double *));
    dVz = (double **) calloc(PageSize,sizeof(double *));
    NoImpacts = (bool **) calloc(PageSize,sizeof(bool *));

//	 4(4) " 3/8" ( ,   4(4) ):
    c4 = (double *) calloc(5,sizeof(double));
    b4 = (double *) calloc(5,sizeof(double));
    a4 = (double **) calloc(5,sizeof(double *));
    for (int i=0;i<5;i++) a4[i] = (double *) calloc(4,sizeof(double));
    c4[2]=1/3.0;  a4[2][1]=1/3.0;
    c4[3]=2/3.0;  a4[3][1]=-1/3.0;  a4[3][2]=1;
    c4[4]=1;      a4[4][1]=1;       a4[4][2]=-1;   a4[4][3]=1;
                  b4[1]=1/8.0;      b4[2]=3/8.0;   b4[3]=3/8.0;  b4[4]=1/8.0;

//	 5(6) -:
    c5 = (double *) calloc(7,sizeof(double));
    b5 = (double *) calloc(7,sizeof(double));
    a5 = (double **) calloc(7,sizeof(double *));
    for (int i=0;i<7;i++) a5[i] = (double *) calloc(6,sizeof(double));
    c5[2]=1/3.0; a5[2][1]=1/3.0;
    c5[3]=2/5.0; a5[3][1]=4/25.0; a5[3][2]=6/25.0;
    c5[4]=1;     a5[4][1]=1/4.0;  a5[4][2]=-3;      a5[4][3]=15/4.0;
    c5[5]=2/3.0; a5[5][1]=6/81.0; a5[5][2]=90/81.0; a5[5][3]=-50/81.0; a5[5][4]=8/81.0;
    c5[6]=4/5.0; a5[6][1]=6/75.0; a5[6][2]=36/75.0; a5[6][3]=10/75.0;  a5[6][4]=8/75.0; a5[6][5]=0;
                 b5[1]=23/192.0;  b5[2]=0;          b5[3]=125/192.0;   b5[4]=0;         b5[5]=-81/192.0; b5[6]=125/192.0;

//	 6(7) :
    c6 = (double *) calloc(8,sizeof(double));
    b6 = (double *) calloc(8,sizeof(double));
    a6 = (double **) calloc(8,sizeof(double *));
    for (int i=0;i<8;i++) a6[i] = (double *) calloc(7,sizeof(double));
    c6[2]=0.5;   a6[2][1]= 0.5;
    c6[3]=2/3.0; a6[3][1]= 2/9.0;    a6[3][2]= 4/9.0;
    c6[4]=1/3.0; a6[4][1]= 7/36.0;   a6[4][2]= 2/9.0;   a6[4][3]=-1/12.0;
    c6[5]=5/6.0; a6[5][1]=-35/144.0; a6[5][2]=-55/36.0; a6[5][3]= 35/48.0;  a6[5][4]= 15/8.0;
    c6[6]=1/6.0; a6[6][1]=-1/360.0;  a6[6][2]=-11/36.0; a6[6][3]=-1/8.0;    a6[6][4]= 0.5;      a6[6][5]=0.1;
    c6[7]=1;     a6[7][1]=-41/260.0; a6[7][2]= 22/13.0; a6[7][3]= 43/156.0; a6[7][4]=-118/39.0; a6[7][5]=32/195.0; a6[7][6]=80/39.0;
                 b6[1]=13/200.0;     b6[2]=0;           b6[3]=11/40.0;      b6[4]=11/40.0;      b6[5]=4/25.0;      b6[6]=4/25.0;     b6[7]=13/200.0;

//	 7(11) :
    c7 = (double *) calloc(12,sizeof(double));
    b7 = (double *) calloc(12,sizeof(double));
    a7 = (double **) calloc(12,sizeof(double *));
    for (int i=0;i<12;i++) a7[i] = (double *) calloc(11,sizeof(double));
    c7[2]=2/27.0; a7[2][1]=2/27.0;
    c7[3]=1/9.0;  a7[3][1]=1/36.0;        a7[3][2]=1/12.0;
    c7[4]=1/6.0;  a7[4][1]=1/24.0;        a7[4][2]=0;      a7[4][3]=1/8.0;
    c7[5]=5/12.0; a7[5][1]=5/12.0;        a7[5][2]=0;      a7[5][3]=-25/16.0; a7[5][4]=25/16.0;
    c7[6]=0.5;    a7[6][1]=1/20.0;        a7[6][2]=0;      a7[6][3]=0;        a7[6][4]=1/4.0;       a7[6][5]=1/5.0;
    c7[7]=5/6.0;  a7[7][1]=-25/108.0;     a7[7][2]=0;      a7[7][3]=0;        a7[7][4]=125/108.0;   a7[7][5]=-65/27.0;     a7[7][6]=125/54.0;
    c7[8]=1/6.0;  a7[8][1]=31/300.0;      a7[8][2]=0;      a7[8][3]=0;        a7[8][4]=0;           a7[8][5]=61/225.0;     a7[8][6]=-2/9.0;     a7[8][7]=13/900.0;
    c7[9]=2/3.0;  a7[9][1]=2;             a7[9][2]=0;      a7[9][3]=0;        a7[9][4]=-53/6.0;     a7[9][5]=704/45.0;     a7[9][6]=-107/9.0;   a7[9][7]=67/90.0;      a7[9][8]=3;
    c7[10]=1/3.0; a7[10][1]=-91/108.0;    a7[10][2]=0;     a7[10][3]=0;       a7[10][4]=23/108.0;   a7[10][5]=-976/135.0;  a7[10][6]=311/54.0;  a7[10][7]=-19/60.0;    a7[10][8]=17/6.0;  a7[10][9]=-1/12.0;
    c7[11]=1;     a7[11][1]=2383/4100.0;  a7[11][2]=0;     a7[11][3]=0;       a7[11][4]=-341/164.0; a7[11][5]=4496/1025.0; a7[11][6]=-301/82.0; a7[11][7]=2133/4100.0; a7[11][8]=45/82.0; a7[11][9]=45/164.0; a7[11][10]=18/41.0;
                  b7[1]=41/840.0;         b7[2]=0;         b7[3]=0;           b7[4]=0;              b7[5]=0;               b7[6]=34/105.0;      b7[7]=9/35.0;          b7[8]=9/35.0;      b7[9]=9/280.0;      b7[10]=9/280.0;     b7[11]=41/840.0;

//	 8(12) -:
    c8 = (double *) calloc(13,sizeof(double));
    b8 = (double *) calloc(13,sizeof(double));
    a8 = (double **) calloc(13,sizeof(double *));
    for (int i=0;i<13;i++) a8[i] = (double *) calloc(12,sizeof(double));
/*
    c8[2]=1/18.0;                     a8[2][1]=1/18.0;
    c8[3]=1/12.0;                     a8[3][1]=1/48.0;                     a8[3][2]=1/16.0;
    c8[4]=1/8.0;                      a8[4][1]=1/32.0;                     a8[4][2]=0;      a8[4][3]=3/32.0;
    c8[5]=5/16.0;                     a8[5][1]=5/16.0;                     a8[5][2]=0;      a8[5][3]=-75/64.0;  a8[5][4]=75/64.0;
    c8[6]=3/8.0;                      a8[6][1]=3/80.0;                     a8[6][2]=0;      a8[6][3]=0;         a8[6][4]=3/16.0;                        a8[6][5]=3/20.0;
    c8[7]=59/400.0;                   a8[7][1]=29443841.0/614563906.0;     a8[7][2]=0;      a8[7][3]=0;         a8[7][4]=77736538.0/692538347.0;        a8[7][5]=-28693883.0/1125000000.0;   a8[7][6]=23124283.0/1800000000.0;
    c8[8]=93/200.0;                   a8[8][1]=16016141.0/946692911.0;     a8[8][2]=0;      a8[8][3]=0;         a8[8][4]=61564180.0/158732637.0;        a8[8][5]=22789713.0/633445777.0;     a8[8][6]=545815736.0/2771057229.0;     a8[8][7]=-180193667.0/1043307555.0;
    c8[9]=5490023248.0/9719169821.0;  a8[9][1]=39632708.0/573591083.0;     a8[9][2]=0;      a8[9][3]=0;         a8[9][4]=-433636366.0/683701615.0;      a8[9][5]=-421739975.0/2616292301.0;  a8[9][6]=100302831.0/723423059.0;      a8[9][7]=790204164.0/839813087.0;      a8[9][8]=800635310.0/3783071287.0;
    c8[10]=13/20.0;                   a8[10][1]=246121993.0/1340847787.0;  a8[10][2]=0;     a8[10][3]=0;        a8[10][4]=-37695042795.0/15268766246.0; a8[10][5]=-309121744.0/1061227803.0; a8[10][6]=-12992083.0/490766935.0;     a8[10][7]=6005943493.0/2108947869.0;   a8[10][8]=393006217.0/1396673457.0;    a8[10][9]=123872331.0/1001029789.0;
    c8[11]=1201146811.0/1299019798.0; a8[11][1]=-1028468189.0/846180014.0; a8[11][2]=0;     a8[11][3]=0;        a8[11][4]=8478235783.0/508512852.0;     a8[11][5]=1311729495.0/1432422823.0; a8[11][6]=-10304129995.0/1701304382.0; a8[11][7]=-48777925059.0/3047939560.0; a8[11][8]=15336726248.0/1032824649.0;  a8[11][9]=-45442868181.0/3398467696.0; a8[11][10]=3065993473.0/597172653.0;
    c8[12]=1;                         a8[12][1]=185892177.0/718116043.0;   a8[12][2]=0;     a8[12][3]=0;        a8[12][4]=-3185094517.0/667107341.0;    a8[12][5]=-477755414.0/1098053517.0; a8[12][6]=-703635378.0/230739211.0;    a8[12][7]=5731566787.0/1027545527.0;   a8[12][8]=5232866602.0/850066563.0;    a8[12][9]=-4093664535.0/808688257.0;   a8[12][10]=3962137247.0/1805957418.0; a8[12][11]=65686358.0/487910083.0;
    c8[13]=1;                         a8[13][1]=403863854.0/491063109.0;   a8[13][2]=0;     a8[13][3]=0;        a8[13][4]=-5068492393.0/434740067.0;    a8[13][5]=-411421997.0/5430433805.0; a8[13][6]=652783627.0/914296604.0;     a8[13][7]=11173962825.0/925320556.0;   a8[13][8]=-13158990841.0/6184727034.0; a8[13][9]=3936647629.0/1978049680.0;   a8[13][10]=-160528059.0/685178525.0;  a8[13][11]=248638103.0/1413531060.0; a8[13][12]=0;
                                      b8[1]=14005451.0/335480064.0;        b8[2]=0;         b8[3]=0;            b8[4]=0;                                b8[5]=0;                             b8[6]=-59238493.0/1068277825.0;        b8[7]=181606767.0/758867731.0;         b8[8]=561292985.0/797845732.0;         b8[9]=-1041891430.0/1371343529.0;      b8[10]=760417239.0/1151165299.0;      b8[11]=118820643.0/751138087.0;      b8[12]=-528747749.0/2220607170.0; b8[13]=1/4.0;

*/
	double	c_2, c_3, c_4, c_5, c_6, c_7, c_8, c_9, c_10, c_11;
	double	b_1, b_6, b_7, b_8, b_9, b_10, b_11, b_12;
	double	a21, a31, a32, a41, a43, a51, a53, a54, a61, a64, a65, a71, a74, a75, a76;
	double	a81, a84, a85, a86, a87, a91, a94, a95, a96, a97, a98;
	double	a101, a104, a105, a106, a107, a108, a109;
	double	a111, a114, a115, a116, a117, a118, a119, a1110;
	double	a121, a124, a125, a126, a127, a128, a129, a1210, a1211;

	c_2  = 0.526001519587677318785587544488E-01;
	c_3  = 0.789002279381515978178381316732E-01;
	c_4  = 0.118350341907227396726757197510E+00;
	c_5  = 0.281649658092772603273242802490E+00;
	c_6  = 0.333333333333333333333333333333E+00;
	c_7  = 0.25E+00;
	c_8  = 0.307692307692307692307692307692E+00;
	c_9  = 0.651282051282051282051282051282E+00;
	c_10 = 0.6E+00;
	c_11 = 0.857142857142857142857142857142E+00;

	b_1 =   5.42937341165687622380535766363E-2;
	b_6 =   4.45031289275240888144113950566E0;
	b_7 =   1.89151789931450038304281599044E0;
	b_8 =  -5.8012039600105847814672114227E0;
	b_9 =   3.1116436695781989440891606237E-1;
	b_10 = -1.52160949662516078556178806805E-1;
	b_11 =  2.01365400804030348374776537501E-1;
	b_12 =  4.47106157277725905176885569043E-2;

	a21 =    5.26001519587677318785587544488E-2;
    a31 =    1.97250569845378994544595329183E-2;
    a32 =    5.91751709536136983633785987549E-2;
    a41 =    2.95875854768068491816892993775E-2;
    a43 =    8.87627564304205475450678981324E-2;
    a51 =    2.41365134159266685502369798665E-1;
    a53 =   -8.84549479328286085344864962717E-1;
    a54 =    9.24834003261792003115737966543E-1;
    a61 =    3.7037037037037037037037037037E-2;
    a64 =    1.70828608729473871279604482173E-1;
    a65 =    1.25467687566822425016691814123E-1;
    a71 =    3.7109375E-2;
    a74 =    1.70252211019544039314978060272E-1;
    a75 =    6.02165389804559606850219397283E-2;
    a76 =   -1.7578125E-2;
    a81 =    3.70920001185047927108779319836E-2;
    a84 =    1.70383925712239993810214054705E-1;
    a85 =    1.07262030446373284651809199168E-1;
    a86 =   -1.53194377486244017527936158236E-2;
    a87 =    8.27378916381402288758473766002E-3;
    a91 =    6.24110958716075717114429577812E-1;
    a94 =   -3.36089262944694129406857109825E0;
    a95 =   -8.68219346841726006818189891453E-1;
    a96 =    2.75920996994467083049415600797E1;
    a97 =    2.01540675504778934086186788979E1;
    a98 =   -4.34898841810699588477366255144E1;
    a101 =   4.77662536438264365890433908527E-1;
    a104 =  -2.48811461997166764192642586468E0;
    a105 =  -5.90290826836842996371446475743E-1;
    a106 =   2.12300514481811942347288949897E1;
    a107 =   1.52792336328824235832596922938E1;
    a108 =  -3.32882109689848629194453265587E1;
    a109 =  -2.03312017085086261358222928593E-2;
    a111 =  -9.3714243008598732571704021658E-1;
    a114 =   5.18637242884406370830023853209E0;
    a115 =   1.09143734899672957818500254654E0;
    a116 =  -8.14978701074692612513997267357E0;
    a117 =  -1.85200656599969598641566180701E1;
    a118 =   2.27394870993505042818970056734E1;
    a119 =   2.49360555267965238987089396762E0;
    a1110 = -3.0467644718982195003823669022E0;
    a121 =   2.27331014751653820792359768449E0;
    a124 =  -1.05344954667372501984066689879E1;
    a125 =  -2.00087205822486249909675718444E0;
    a126 =  -1.79589318631187989172765950534E1;
    a127 =   2.79488845294199600508499808837E1;
    a128 =  -2.85899827713502369474065508674E0;
    a129 =  -8.87285693353062954433549289258E0;
    a1210 =  1.23605671757943030647266201528E1;
    a1211 =  6.43392746015763530355970484046E-1;

    c8[2]=c_2;    a8[2][1]=a21;
    c8[3]=c_3;    a8[3][1]=a31;    a8[3][2]=a32;
    c8[4]=c_4;    a8[4][1]=a41;    a8[4][2]=0;   a8[4][3]=a43;
    c8[5]=c_5;    a8[5][1]=a51;    a8[5][2]=0;   a8[5][3]=a53;  a8[5][4]=a54;
    c8[6]=c_6;    a8[6][1]=a61;    a8[6][2]=0;   a8[6][3]=0;    a8[6][4]=a64;    a8[6][5]=a65;
    c8[7]=c_7;    a8[7][1]=a71;    a8[7][2]=0;   a8[7][3]=0;    a8[7][4]=a74;    a8[7][5]=a75;    a8[7][6]=a76;
    c8[8]=c_8;    a8[8][1]=a81;    a8[8][2]=0;   a8[8][3]=0;    a8[8][4]=a84;    a8[8][5]=a85;    a8[8][6]=a86;    a8[8][7]=a87;
    c8[9]=c_9;    a8[9][1]=a91;    a8[9][2]=0;   a8[9][3]=0;    a8[9][4]=a94;    a8[9][5]=a95;    a8[9][6]=a96;    a8[9][7]=a97;    a8[9][8]=a98;
    c8[10]=c_10;  a8[10][1]=a101;  a8[10][2]=0;  a8[10][3]=0;   a8[10][4]=a104;  a8[10][5]=a105;  a8[10][6]=a106;  a8[10][7]=a107;  a8[10][8]=a108;  a8[10][9]=a109;
    c8[11]=c_11;  a8[11][1]=a111;  a8[11][2]=0;  a8[11][3]=0;   a8[11][4]=a114;  a8[11][5]=a115;  a8[11][6]=a116;  a8[11][7]=a117;  a8[11][8]=a118;  a8[11][9]=a119;  a8[11][10]=a1110;
    c8[12]=0;     a8[12][1]=a121;  a8[12][2]=0;  a8[12][3]=0;   a8[12][4]=a124;  a8[12][5]=a125;  a8[12][6]=a126;  a8[12][7]=a127;  a8[12][8]=a128;  a8[12][9]=a129;  a8[12][10]=a1210;  a8[12][11]=a1211;
                  b8[1]=b_1;       b8[2]=0;      b8[3]=0;       b8[4]=0;         b8[5]=0;         b8[6]=b_6;       b8[7]=b_7;       b8[8]=b_8;       b8[9]=b_9;       b8[10]=b_10;       b8[11]=b_11;       b8[12]=b_12;

    Camera = new CAMERA;
    Time = new TTimeEx;
    BMP = new Byte*[4096];
    Pages = 1;
    Count = 0;
    Tau = 0;
	Active = -1;
    Selected = -1;
    Design = true;
    Draw3D = false;
	Brightness = 3;
    Image = NULL;
    Gamma = 10;
    DetectImpacts = 1;
    Alpha = 0.5;
    Method = 2;
    Color = clWhite;
};

// ----------------------------------------------------------------------------
// :

TUniverse::~TUniverse()
{
	int i;
    for (i=0; i<Count; i++) {
    	delete Planets[i];
        free(dx[i]);  free(dy[i]);  free(dz[i]);
        free(dVx[i]); free(dVy[i]); free(dVz[i]);
        free(NoImpacts[i]);
    }
    free(Planets);
    free(Numbers);  free(Criteria);
    free(dx);  free(dy);  free(dz);
    free(dVx); free(dVy); free(dVz);
    free(NoImpacts);  free(BMP);
    for (i=0; i<5; i++) free(a4[i]); free(a4);
    for (i=0; i<7; i++) free(a5[i]); free(a5);
    for (i=0; i<8; i++) free(a6[i]); free(a6);
    for (i=0; i<12; i++) free(a7[i]); free(a7);
    for (i=0; i<13; i++) free(a8[i]); free(a8);
//    delete Camera;
//    delete Time;
};

// ----------------------------------------------------------------------------
//   3D-:
bool TUniverse::GetDraw3D()
{
	return Draw3D;
};
// ----------------------------------------------------------------------------
//   3D-:
void TUniverse::SetDraw3D(bool d)
{
	Draw3D = d;
    if (Draw3D)  if (Image!=NULL) Image->Picture->Bitmap->PixelFormat = pf24bit;
    if (!Draw3D) if (Image!=NULL) Image->Picture->Bitmap->PixelFormat = pfDevice;
};
// ----------------------------------------------------------------------------
//     .

int TUniverse::NewPlanet(TColor PColor,double Radius,double Mass,double X,double Y,double Z,double Vx,double Vy,double Vz,bool CheckValid)
{
//	 :
	TPlanet *New = new TPlanet(Image,PColor,Radius,Mass,X,Y,Z,Vx,Vy,Vz);
    New->EdgeColor = TColor(Color & clBlack);
    New->ActColor  = TColor(Color & clBlue);
    New->TrajColor = TColor(Color & PColor);
//	   :
	if (CheckValid)
		if (!CheckPlanet(New,-1)) {
    		delete New;
        	return 0;
    	}
//	    :
	Planets[Count] = New;
    dx[Count]  = (double *) calloc(NRegs,sizeof(double));
    dy[Count]  = (double *) calloc(NRegs,sizeof(double));
    dz[Count]  = (double *) calloc(NRegs,sizeof(double));
    dVx[Count] = (double *) calloc(NRegs,sizeof(double));
    dVy[Count] = (double *) calloc(NRegs,sizeof(double));
    dVz[Count] = (double *) calloc(NRegs,sizeof(double));
    NoImpacts[Count] = (bool *) calloc(PageSize*Pages,sizeof(bool));
    Count++;
    if (Pages*PageSize == Count) {
    	Pages++;
    	Planets = (TPlanet **)realloc(Planets,  PageSize*Pages*sizeof(TPlanet *));
    	Numbers =  (int *)    realloc(Numbers,  PageSize*Pages*sizeof(int));
	    Criteria = (double *) realloc(Criteria, PageSize*Pages*sizeof(double));
        dx  = (double **) realloc(dx, PageSize*Pages*sizeof(double *));
        dy  = (double **) realloc(dy, PageSize*Pages*sizeof(double *));
        dz  = (double **) realloc(dz, PageSize*Pages*sizeof(double *));
        dVx = (double **) realloc(dVx, PageSize*Pages*sizeof(double *));
        dVy = (double **) realloc(dVy, PageSize*Pages*sizeof(double *));
        dVz = (double **) realloc(dVz, PageSize*Pages*sizeof(double *));
        NoImpacts = (bool **) realloc(NoImpacts, PageSize*Pages*sizeof(bool *));
        for (int i=0; i<Count; i++)
        	NoImpacts[i] = (bool *) realloc(NoImpacts[i], PageSize*Pages*sizeof(bool));
    }
    return 1;
};

// ----------------------------------------------------------------------------
//         Fx  .
//        ,
//  0.   ,   
//      RandomPlanets.
//    r0  r1,   AutoRadius.

int TUniverse::NewPlanet(double Fx, bool AutoRadius, double r0, double r1)
{
//	 :
	TPlanet *New = new TPlanet(Image,Camera,Fx);
    New->EdgeColor = TColor(Color & clBlack);
    New->ActColor  = TColor(Color & clBlue);
    New->TrajColor = TColor(Color & (New->Color));
    if (!AutoRadius) {
    	New->Radius = trunc(r0 + (r1-r0)*double(rand())/RAND_MAX);
        New->Mass = (New->Radius)*(New->Radius)*(New->Radius);
    }
//	   :
	if (!CheckPlanet(New,-1)) {
    	delete New;
        return 0;
    }
//	    :
	Planets[Count] = New;
    dx[Count]  = (double *) calloc(NRegs,sizeof(double));
    dy[Count]  = (double *) calloc(NRegs,sizeof(double));
    dz[Count]  = (double *) calloc(NRegs,sizeof(double));
    dVx[Count] = (double *) calloc(NRegs,sizeof(double));
    dVy[Count] = (double *) calloc(NRegs,sizeof(double));
    dVz[Count] = (double *) calloc(NRegs,sizeof(double));
    NoImpacts[Count] = (bool *) calloc(PageSize*Pages,sizeof(bool));
    Count++;
    if (Pages*PageSize == Count) {
    	Pages++;
    	Planets = (TPlanet **) realloc(Planets, PageSize*Pages*sizeof(TPlanet *));
    	Numbers =  (int *)    realloc(Numbers,  PageSize*Pages*sizeof(int));
	    Criteria = (double *) realloc(Criteria, PageSize*Pages*sizeof(double));
        dx  = (double **) realloc(dx, PageSize*Pages*sizeof(double *));
        dy  = (double **) realloc(dy, PageSize*Pages*sizeof(double *));
        dz  = (double **) realloc(dz, PageSize*Pages*sizeof(double *));
        dVx = (double **) realloc(dVx, PageSize*Pages*sizeof(double *));
        dVy = (double **) realloc(dVy, PageSize*Pages*sizeof(double *));
        dVz = (double **) realloc(dVz, PageSize*Pages*sizeof(double *));
        NoImpacts = (bool **) realloc(NoImpacts, PageSize*Pages*sizeof(bool *));
        for (int i=0; i<Count; i++)
        	NoImpacts[i] = (bool *) realloc(NoImpacts[i], PageSize*Pages*sizeof(bool));
    }
    return 1;
};

// ----------------------------------------------------------------------------
//  N  .     ,
//    1  .     
//      ,    2,3,...
//  .   AutoLimits,     
//    .

void TUniverse::RandomPlanets(int N, bool AutoLimits, double r0, double r1, double Vx0, double Vx1,
                              double Vy0, double Vy1, double Vz0, double Vz1)
{
    double Fx = 1;							//    
	int i,j,ok;

    for (i=0; i<N; i++) {
    	j=0;
    	do {
        	//     ,  :
        	if (j>=10000) {Fx++; j=0;}
            // :
        	ok = NewPlanet(Fx,AutoLimits,r0,r1);
            //  ,    ( ):
            if (ok && !AutoLimits) {
            	Planets[Count-1]->Vx = Vx0 + (Vx1-Vx0)*double(rand())/RAND_MAX;
            	Planets[Count-1]->Vy = Vy0 + (Vy1-Vy0)*double(rand())/RAND_MAX;
            	Planets[Count-1]->Vz = Vz0 + (Vz1-Vz0)*double(rand())/RAND_MAX;
            }
            j++;
        } while (!ok);
    }
};
// ----------------------------------------------------------------------------
//  N     .

void TUniverse::RandomPlanets(int N, double r0, double r1, double x0, double x1, double y0, double y1, double z0, double z1,
                             double Vx0, double Vx1, double Vy0, double Vy1, double Vz0, double Vz1)
{
    double r,x,y,z,Vx,Vy,Vz;
	int i,j,ok;

    for (i=0; i<N; i++) {
    	j=0;
    	do {
        	//     ,  :
        	if (j>=10000) {
            	x0 -= 0.5*(x1-x0);  x1 += 0.5*(x1-x0);
            	y0 -= 0.5*(y1-y0);  y1 += 0.5*(y1-y0);
            	z0 -= 0.5*(z1-z0);  x1 += 0.5*(z1-z0);
                j = 0;
            }
            // :
            r = trunc(r0 + (r1-r0)*double(rand())/RAND_MAX);
            x = x0 + (x1-x0)*double(rand())/RAND_MAX;
            y = y0 + (y1-y0)*double(rand())/RAND_MAX;
            z = z0 + (z1-z0)*double(rand())/RAND_MAX;
            Vx = Vx0 + (Vx1-Vx0)*double(rand())/RAND_MAX;
            Vy = Vy0 + (Vy1-Vy0)*double(rand())/RAND_MAX;
            Vz = Vz0 + (Vz1-Vz0)*double(rand())/RAND_MAX;
        	ok = NewPlanet(TColor(random(256)*256*256 + random(256)*256 + random(256)),
                           r, r*r*r, x,y,z, Vx,Vy,Vz, true);
            j++;
        } while (!ok);
    }
};

// ----------------------------------------------------------------------------
//  :

void TUniverse::DeletePlanet(int Del)
{
	int i;
//	:
	if (Del<0 || Del>=Count) return;
    delete Planets[Del];
    free(dx[Count-1]);  free(dy[Count-1]);  free(dz[Count-1]);
    free(dVx[Count-1]); free(dVy[Count-1]); free(dVz[Count-1]);
    free(NoImpacts[Count-1]);
//	    :
	for (i=Del+1; i<Count; i++)
    	Planets[i-1] = Planets[i];
//	 :
    Count--;
	if ((Pages-1)*PageSize > Count) {
    	Pages--;
    	Planets = (TPlanet **) realloc(Planets, PageSize*Pages*sizeof(TPlanet *));
    	Numbers =  (int *)    realloc(Numbers,  PageSize*Pages*sizeof(int));
	    Criteria = (double *) realloc(Criteria, PageSize*Pages*sizeof(double));
        dx  = (double **) realloc(dx, PageSize*Pages*sizeof(double *));
        dy  = (double **) realloc(dy, PageSize*Pages*sizeof(double *));
        dz  = (double **) realloc(dz, PageSize*Pages*sizeof(double *));
        dVx = (double **) realloc(dVx, PageSize*Pages*sizeof(double *));
        dVy = (double **) realloc(dVy, PageSize*Pages*sizeof(double *));
        dVz = (double **) realloc(dVz, PageSize*Pages*sizeof(double *));
        NoImpacts = (bool **) realloc(NoImpacts, PageSize*Pages*sizeof(bool *));
        for (i=0; i<Count; i++)
        	NoImpacts[i] = (bool *) realloc(NoImpacts[i], PageSize*Pages*sizeof(bool));
    }
//	   :
    if (Active > Del) Active--;
    else if (Active == Del) Active = -1;
//	   :
    if (Selected > Del) Selected--;
    else if (Selected == Del) Selected = -1;
};

// ----------------------------------------------------------------------------
//   :

void TUniverse::Init()
{
    for (int i=0; i<Count; i++) {
    	delete Planets[i];
        free(dx[i]);  free(dy[i]);  free(dz[i]);
        free(dVx[i]); free(dVy[i]); free(dVz[i]);
        free(NoImpacts[i]);
    }
    free(Planets);
    free(dx);  free(dy);  free(dz);
    free(dVx); free(dVy); free(dVz);
    free(NoImpacts);

    Planets = (TPlanet **) calloc(PageSize,sizeof(TPlanet *));
    Numbers =  (int *)    calloc(PageSize,sizeof(int));
    Criteria = (double *) calloc(PageSize,sizeof(double));
    dx  = (double **) calloc(PageSize,sizeof(double *));
    dy  = (double **) calloc(PageSize,sizeof(double *));
    dz  = (double **) calloc(PageSize,sizeof(double *));
    dVx = (double **) calloc(PageSize,sizeof(double *));
    dVy = (double **) calloc(PageSize,sizeof(double *));
    dVz = (double **) calloc(PageSize,sizeof(double *));
    NoImpacts = (bool **) calloc(PageSize,sizeof(bool *));
    Pages = 1;
    Count = 0;
    Active = -1;
    Selected = -1;
};

// ----------------------------------------------------------------------------
// :

void TUniverse::Draw()
{
	if (Image==NULL) return;

    //  :
    Image->Canvas->Pen->Mode    = pmCopy;
    Image->Canvas->Pen->Color   = Color;
    Image->Canvas->Brush->Color = Color;
    Image->Canvas->Brush->Style = bsSolid;
    Image->Canvas->Rectangle(0,0,Image->Width,Image->Height);

    //    :
    double **R = Camera->R;					//  
    int i, Count2 = 0;
    double X0 = Camera->X0;
    double Y0 = Camera->Y0;
    double Z0 = Camera->Z0;
    double XR,YR,ZR;						//   . 
	for (i=0; i<Count; i++) {				//   :
    	ZR = R[2][0]*(Planets[i]->X-X0) + R[2][1]*(Planets[i]->Y-Y0) + R[2][2]*(Planets[i]->Z-Z0);
        if (ZR >= Planets[i]->Radius) continue;
    	XR = R[0][0]*(Planets[i]->X-X0) + R[0][1]*(Planets[i]->Y-Y0) + R[0][2]*(Planets[i]->Z-Z0);
    	YR = R[1][0]*(Planets[i]->X-X0) + R[1][1]*(Planets[i]->Y-Y0) + R[1][2]*(Planets[i]->Z-Z0);
        Numbers[Count2] = i;
        Criteria[Count2] = XR*XR + YR*YR + ZR*ZR - (Planets[i]->Radius)*(Planets[i]->Radius);
        Count2++;
    }

    //     :
    if (Count2 > 0) {
    	int *Order = TreeSort(Numbers, Criteria, Count2);	// 
		for (i=0; i<Count2; i++) 							// 
    		Planets[Order[i]]->Draw(Camera,Design,Draw3D,Brightness,BMP);
        delete[] Order;
    }

    //  :
    for (i=0; i<Count; i++) Planets[i]->DrawTrajectory(Camera);

	//   .
    if (Camera->DrawAxes) {
	    //  :
	    int L = 24;  int Lt = (5*L)/2;
	    int x0 = Image->Width - 3*L;
	    int y0 = 3*L;
	    //   :
	    double ix = R[0][0];  double jx = R[0][1];  double kx = R[0][2];
	    double iy = R[1][0];  double jy = R[1][1];  double ky = R[1][2];
	    double iz = R[2][0];  double jz = R[2][1];  double kz = R[2][2];
	    //    :
	    double x1,y1,x2,y2,x3,y3;
	    if (iz <= jz && jz <= kz) {x1 = ix; x2 = jx; x3 = kx;
	                               y1 = iy; y2 = jy; y3 = ky;}
	    if (iz <= kz && kz <= jz) {x1 = ix; x2 = kx; x3 = jx;
	                               y1 = iy; y2 = ky; y3 = jy;}
	    if (jz <= iz && iz <= kz) {x1 = jx; x2 = ix; x3 = kx;
	                               y1 = jy; y2 = iy; y3 = ky;}
	    if (jz <= kz && kz <= iz) {x1 = jx; x2 = kx; x3 = ix;
	                               y1 = jy; y2 = ky; y3 = iy;}
	    if (kz <= iz && iz <= jz) {x1 = kx; x2 = ix; x3 = jx;
	                               y1 = ky; y2 = iy; y3 = jy;}
	    if (kz <= jz && jz <= iz) {x1 = kx; x2 = jx; x3 = ix;
	                               y1 = ky; y2 = jy; y3 = iy;}
	    //    :
	    Image->Canvas->Pen->Mode = pmCopy;
	    Image->Canvas->Pen->Color = clNavy;
	    Image->Canvas->Brush->Style = bsSolid;
	    Image->Canvas->Brush->Color = Color;
	    TPoint Points[3];

	    Image->Canvas->MoveTo(x0, y0);
	    Image->Canvas->LineTo(x0 + 2*L*x1, y0 - 2*L*y1);
	    Points[0] = Point(x0,y0);
	    Points[1] = Point(x0+L*x1, y0-L*y1);
	    Points[2] = Point(x0+L*x2, y0-L*y2);
	    Image->Canvas->Polygon(Points, 2);

	    Points[0] = Point(x0,y0);
	    Points[1] = Point(x0+L*x1, y0-L*y1);
	    Points[2] = Point(x0+L*x3, y0-L*y3);
	    Image->Canvas->Polygon(Points, 2);

	    Image->Canvas->MoveTo(x0 + 2*L*x2, y0 - 2*L*y2);
	    Image->Canvas->LineTo(x0, y0);
	    Image->Canvas->LineTo(x0 + 2*L*x3, y0 - 2*L*y3);
	    Points[0] = Point(x0,y0);
	    Points[1] = Point(x0+L*x2, y0-L*y2);
	    Points[2] = Point(x0+L*x3, y0-L*y3);
	    Image->Canvas->Polygon(Points, 2);
	    //  :
	    int w,h;
	    Image->Canvas->Brush->Style = bsClear;
	    w = Image->Canvas->TextWidth("x");
	    h = Image->Canvas->TextHeight("x");
	    Image->Canvas->TextOut(x0 + Lt*ix - w/2, y0 - Lt*iy - h/2, "x");
	    w = Image->Canvas->TextWidth("y");
	    h = Image->Canvas->TextHeight("y");
	    Image->Canvas->TextOut(x0 + Lt*jx - w/2, y0 - Lt*jy - h/2, "y");
	    w = Image->Canvas->TextWidth("z");
	    h = Image->Canvas->TextHeight("z");
	    Image->Canvas->TextOut(x0 + Lt*kx - w/2, y0 - Lt*ky - h/2, "z");
    }

    //  .
    if (Camera->DrawPoint) {
    	int s = 10;
        int h = Image->Canvas->TextHeight("0");
	    Image->Canvas->Brush->Style = bsClear;
     	Image->Canvas->TextOut(s,s,       "x = " + Float2String(Camera->X0,9));
     	Image->Canvas->TextOut(s,s + h,   "y = " + Float2String(Camera->Y0,9));
     	Image->Canvas->TextOut(s,s + 2*h, "z = " + Float2String(Camera->Z0,9));
    }
};

// ----------------------------------------------------------------------------
//   .     
//    .    0.
//     , -1.

int TUniverse::Touched(int MX, int MY)
{
    if (Image==NULL) return -1;

    //    :
    double **R = Camera->R;					//  
    int i, Count2 = 0, i0 = -1;
    double X0 = Camera->X0;
    double Y0 = Camera->Y0;
    double Z0 = Camera->Z0;
    double XR,YR,ZR;						//   . 
	for (i=0; i<Count; i++) {				//   :
    	ZR = R[2][0]*(Planets[i]->X-X0) + R[2][1]*(Planets[i]->Y-Y0) + R[2][2]*(Planets[i]->Z-Z0);
        if (-ZR <= Planets[i]->Radius) continue;
    	XR = R[0][0]*(Planets[i]->X-X0) + R[0][1]*(Planets[i]->Y-Y0) + R[0][2]*(Planets[i]->Z-Z0);
    	YR = R[1][0]*(Planets[i]->X-X0) + R[1][1]*(Planets[i]->Y-Y0) + R[1][2]*(Planets[i]->Z-Z0);
        Numbers[Count2] = i;
        Criteria[Count2] = XR*XR + YR*YR + ZR*ZR - (Planets[i]->Radius)*(Planets[i]->Radius);
        Count2++;
    }

    //     :
    if (Count2 > 0) {
    	int *Order = TreeSort(Numbers, Criteria, Count2);	// 
		for (i=Count2-1; i>=0; i--)							// 
    		if (Planets[Order[i]]->Touch(MX,MY,Camera,Design)) {
            	i0 = Order[i];
                break;
            }
        delete[] Order;
    }

    return i0;
};

// ----------------------------------------------------------------------------
//   (   ):

void TUniverse::SelectPlanet(int Sel)
{
    if (0<=Sel && Sel<Count) {
    	Planets[Sel]->Selected = !(Planets[Sel]->Selected);
    	Planets[Sel]->DrawSelection(Camera);
    }
	double Xc=0, Yc=0, Zc=0;		//    
    double M = 0;					//  
    for (int i=0; i<Count; i++)
		if (Planets[i]->Selected) {
        	M += Planets[i]->Mass;
            Xc += (Planets[i]->X)*(Planets[i]->Mass);
            Yc += (Planets[i]->Y)*(Planets[i]->Mass);
            Zc += (Planets[i]->Z)*(Planets[i]->Mass);
        }
    if (M > MINDOUBLE) {
    	Xc /= M;  Yc /= M;  Zc /= M;
        Camera->dX0 = Camera->X0 - Xc;
        Camera->dY0 = Camera->Y0 - Yc;
        Camera->dZ0 = Camera->Z0 - Zc;
    }
};

// ----------------------------------------------------------------------------
//     :

void TUniverse::DeselectAll()
{
    int i;
	for (i=0; i<Count; i++)
    	if (Planets[i]->Selected) {
    		Planets[i]->Selected = false;
        	Planets[i]->DrawSelection(Camera);
        }
    Selected = -1;
};

// ----------------------------------------------------------------------------
//   (   ).
//       :

void TUniverse::ActivatePlanet(int Act)
{
//	 :
	if (Active > -1) {
    	Planets[Active]->Active = false;
    	Planets[Active]->DrawActivation(Camera);
    }
//	 :
    if (0<=Act && Act<Count && Act!=Active) {
    	Planets[Act]->Active = true;
    	Planets[Act]->DrawActivation(Camera);
        Active = Act;
    }
    else Active = -1;
};

// ----------------------------------------------------------------------------
//   Planet        ,
//    i0 (     Planet).  i0  ,
//       ( Planet   ).

int TUniverse::CheckPlanet(TPlanet *Planet, int i0)
{
    double X,Y,Z,R;
	for (int i=0; i<Count; i++)
    	if (i!=i0) {
        	X = Planet->X - Planets[i]->X;
            Y = Planet->Y - Planets[i]->Y;
            Z = Planet->Z - Planets[i]->Z;
            R = Planet->Radius + Planets[i]->Radius;
        	if (R*R > X*X + Y*Y + Z*Z) return 0;
        }
    return 1;
};

// ----------------------------------------------------------------------------
//       .  Strict
//   : 1 -    ,  
//     ; 0 -    ,   
//    .

int TUniverse::CheckAllSelected(int Strict)
{
    int k,i;
    double X,Y,Z,R;
	for (k=0; k<Count; k++)
    	if (Planets[k]->Selected)
			for (i=0; i<Count; i++)
		    	if ((Strict && i!=k) || (!Strict && !Planets[i]->Selected)) {
        			X = Planets[k]->X - Planets[i]->X;
            		Y = Planets[k]->Y - Planets[i]->Y;
            		Z = Planets[k]->Z - Planets[i]->Z;
            		R = Planets[k]->Radius + Planets[i]->Radius;
        			if (R*R > X*X + Y*Y + Z*Z) return 0;
                }
    return 1;
};

// ----------------------------------------------------------------------------
//   :

void TUniverse::UpdateSelectRect(TRect *Rect)
{
    int i;
	for (i=0; i<Count; i++)
    	if (Planets[i]->Inside != Planets[i]->InsideRect(Rect,Camera)) {
        	SelectPlanet(i);
            Planets[i]->Inside = !Planets[i]->Inside;
        }
};

// ----------------------------------------------------------------------------
//     -   dt.   
//  Next_ .

void TUniverse::RungeKutta4Step(double dt)
{
	int i,j,k;
    double s, rij, rij3, dX,dY,dZ;

//	      0- :
	for (i=0; i<Count; i++) {
    	dx[i][0]  = Planets[i]->X;		dVx[i][0] = Planets[i]->Vx;
        dy[i][0]  = Planets[i]->Y;		dVy[i][0] = Planets[i]->Vy;
        dz[i][0]  = Planets[i]->Z;		dVz[i][0] = Planets[i]->Vz;
    }
//	 1- :
	for (i=0; i<Count; i++) {
    	dx[i][1]  = dt*dVx[i][0];		dVx[i][1] = 0;
        dy[i][1]  = dt*dVy[i][0];		dVy[i][1] = 0;
        dz[i][1]  = dt*dVz[i][0];		dVz[i][1] = 0;

        for (j=0;j<Count;j++) if (i!=j) {
        	dX = dx[j][0]-dx[i][0];
            dY = dy[j][0]-dy[i][0];
            dZ = dz[j][0]-dz[i][0];
        	rij = sqrt(dX*dX + dY*dY + dZ*dZ);  rij3 = rij*rij*rij;
            dVx[i][1] += (Planets[j]->Mass)*dX / rij3;
            dVy[i][1] += (Planets[j]->Mass)*dY / rij3;
            dVz[i][1] += (Planets[j]->Mass)*dZ / rij3;
        }
        dVx[i][1] *= dt*Gamma;  dVy[i][1] *= dt*Gamma;  dVz[i][1] *= dt*Gamma;
    }
//	 2,3,4 :
    for (k=2; k<=4; k++) {
    	s = (k!=4) ? 0.5 : 1;
		for (i=0; i<Count; i++) {
	    	dx[i][k] = dt*(dVx[i][0] + s*dVx[i][k-1]);
	        dy[i][k] = dt*(dVy[i][0] + s*dVy[i][k-1]);
	        dz[i][k] = dt*(dVz[i][0] + s*dVz[i][k-1]);

	        dVx[i][k] = 0;  dVy[i][k] = 0;  dVz[i][k] = 0;
	        for (j=0;j<Count;j++) if (i!=j) {
            	dX = dx[j][0]-dx[i][0]+s*(dx[j][k-1]-dx[i][k-1]);
                dY = dy[j][0]-dy[i][0]+s*(dy[j][k-1]-dy[i][k-1]);
                dZ = dz[j][0]-dz[i][0]+s*(dz[j][k-1]-dz[i][k-1]);
            	rij = sqrt(dX*dX + dY*dY + dZ*dZ);  rij3 = rij*rij*rij;
                dVx[i][k] += (Planets[j]->Mass)*dX / rij3;
                dVy[i][k] += (Planets[j]->Mass)*dY / rij3;
                dVz[i][k] += (Planets[j]->Mass)*dZ / rij3;
            }
	        dVx[i][k] *= dt*Gamma;  dVy[i][k] *= dt*Gamma;  dVz[i][k] *= dt*Gamma;
	    }
    }
//	     :
	for (i=0; i<Count; i++) {
    	Planets[i]->NextX  = dx[i][0] + (dx[i][1] + 2*dx[i][2] + 2*dx[i][3] + dx[i][4])/6;
        Planets[i]->NextY  = dy[i][0] + (dy[i][1] + 2*dy[i][2] + 2*dy[i][3] + dy[i][4])/6;
        Planets[i]->NextZ  = dz[i][0] + (dz[i][1] + 2*dz[i][2] + 2*dz[i][3] + dz[i][4])/6;
        Planets[i]->NextVx = dVx[i][0] + (dVx[i][1] + 2*dVx[i][2] + 2*dVx[i][3] + dVx[i][4])/6;
        Planets[i]->NextVy = dVy[i][0] + (dVy[i][1] + 2*dVy[i][2] + 2*dVy[i][3] + dVy[i][4])/6;
        Planets[i]->NextVz = dVz[i][0] + (dVz[i][1] + 2*dVz[i][2] + 2*dVz[i][3] + dVz[i][4])/6;
    }
//	     Next_:
    Tau = dt;
//	  :
	if (DetectImpacts) {
		for (i=0; i<Count; i++)
        	for (j=i+1; j<Count; j++)  NoImpacts[i][j] = false;
    }
};

// ----------------------------------------------------------------------------
//      -   dt.   
//  Next_ .

void TUniverse::XRungeKuttaStep(double *c, double *b, double **a, int N, double dt)
{
//	c,b,a -   
//	N -   
	int i,j,k,m;
    double kSUMi, kSUMj, rij, rij3, dX,dY,dZ;

//	      0- :
	for (i=0; i<Count; i++) {
    	dx[i][0]  = Planets[i]->X;		dVx[i][0] = Planets[i]->Vx;
        dy[i][0]  = Planets[i]->Y;		dVy[i][0] = Planets[i]->Vy;
        dz[i][0]  = Planets[i]->Z;		dVz[i][0] = Planets[i]->Vz;
    }
//	 1- :
	for (i=0; i<Count; i++) {
    	dx[i][1]  = dVx[i][0];
        dy[i][1]  = dVy[i][0];
        dz[i][1]  = dVz[i][0];

        dVx[i][1] = 0;  dVy[i][1] = 0;  dVz[i][1] = 0;
        for (j=0;j<Count;j++) if (i!=j) {
        	dX = dx[j][0]-dx[i][0];
            dY = dy[j][0]-dy[i][0];
            dZ = dz[j][0]-dz[i][0];
        	rij = sqrt(dX*dX + dY*dY + dZ*dZ);  rij3 = rij*rij*rij;
            dVx[i][1] += (Planets[j]->Mass)*dX / rij3;
            dVy[i][1] += (Planets[j]->Mass)*dY / rij3;
            dVz[i][1] += (Planets[j]->Mass)*dZ / rij3;
        }
        dVx[i][1] *= Gamma;  dVy[i][1] *= Gamma;  dVz[i][1] *= Gamma;
    }
//	   2-  N-:
    for (k=2; k<=N; k++) {
		for (i=0; i<Count; i++) {
        		kSUMi = 0;  for (m=1;m<k;m++) kSUMi += a[k][m]*dVx[i][m];
	    	dx[i][k] = dVx[i][0] + dt*kSUMi;
        		kSUMi = 0;  for (m=1;m<k;m++) kSUMi += a[k][m]*dVy[i][m];
	        dy[i][k] = dVy[i][0] + dt*kSUMi;
        		kSUMi = 0;  for (m=1;m<k;m++) kSUMi += a[k][m]*dVz[i][m];
	        dz[i][k] = dVz[i][0] + dt*kSUMi;

	        dVx[i][k] = 0;  dVy[i][k] = 0;  dVz[i][k] = 0;
	        for (j=0;j<Count;j++) if (i!=j) {
            	kSUMi = 0; kSUMj = 0;  for (m=1;m<k;m++) { kSUMi += a[k][m]*dx[i][m];
                                                           kSUMj += a[k][m]*dx[j][m]; }
            	dX = dx[j][0]-dx[i][0] + dt*(kSUMj-kSUMi);
            	kSUMi = 0; kSUMj = 0;  for (m=1;m<k;m++) { kSUMi += a[k][m]*dy[i][m];
                                                           kSUMj += a[k][m]*dy[j][m]; }
                dY = dy[j][0]-dy[i][0] + dt*(kSUMj-kSUMi);
            	kSUMi = 0; kSUMj = 0;  for (m=1;m<k;m++) { kSUMi += a[k][m]*dz[i][m];
                                                           kSUMj += a[k][m]*dz[j][m]; }
                dZ = dz[j][0]-dz[i][0] + dt*(kSUMj-kSUMi);
        		rij = sqrt(dX*dX + dY*dY + dZ*dZ);  rij3 = rij*rij*rij;
                dVx[i][k] += (Planets[j]->Mass)*dX / rij3;
                dVy[i][k] += (Planets[j]->Mass)*dY / rij3;
                dVz[i][k] += (Planets[j]->Mass)*dZ / rij3;
            }
	        dVx[i][k] *= Gamma;  dVy[i][k] *= Gamma;  dVz[i][k] *= Gamma;
	    }
    }
//	     :
	for (i=0; i<Count; i++) {
    		kSUMi = 0;  for (m=1;m<=N;m++) kSUMi += b[m]*dx[i][m];
    	Planets[i]->NextX  = dx[i][0] + dt*kSUMi;
    		kSUMi = 0;  for (m=1;m<=N;m++) kSUMi += b[m]*dy[i][m];
        Planets[i]->NextY  = dy[i][0] + dt*kSUMi;
    		kSUMi = 0;  for (m=1;m<=N;m++) kSUMi += b[m]*dz[i][m];
        Planets[i]->NextZ  = dz[i][0] + dt*kSUMi;
    		kSUMi = 0;  for (m=1;m<=N;m++) kSUMi += b[m]*dVx[i][m];
        Planets[i]->NextVx = dVx[i][0] + dt*kSUMi;
    		kSUMi = 0;  for (m=1;m<=N;m++) kSUMi += b[m]*dVy[i][m];
        Planets[i]->NextVy = dVy[i][0] + dt*kSUMi;
    		kSUMi = 0;  for (m=1;m<=N;m++) kSUMi += b[m]*dVz[i][m];
        Planets[i]->NextVz = dVz[i][0] + dt*kSUMi;
    }
//	     Next_:
    Tau = dt;
//	  :
	if (DetectImpacts) {
		for (i=0; i<Count; i++)
        	for (j=i+1; j<Count; j++)  NoImpacts[i][j] = false;
    }
};

// ----------------------------------------------------------------------------
//      2(2)   dt.   
//  Next_ .

void TUniverse::RungeKutta2Step(double dt)
{
	int i,j;
    double rij, rij3, dX,dY,dZ;

//	      0- :
	for (i=0; i<Count; i++) {
    	dx[i][0]  = Planets[i]->X;		dVx[i][0] = Planets[i]->Vx;
        dy[i][0]  = Planets[i]->Y;		dVy[i][0] = Planets[i]->Vy;
        dz[i][0]  = Planets[i]->Z;		dVz[i][0] = Planets[i]->Vz;
    }
//	 1- :
	for (i=0; i<Count; i++) {
    	dx[i][1]  = dt*dVx[i][0];		dVx[i][1] = 0;
        dy[i][1]  = dt*dVy[i][0];		dVy[i][1] = 0;
        dz[i][1]  = dt*dVz[i][0];		dVz[i][1] = 0;

        for (j=0;j<Count;j++) if (i!=j) {
        	dX = dx[j][0]-dx[i][0];
            dY = dy[j][0]-dy[i][0];
            dZ = dz[j][0]-dz[i][0];
        	rij = sqrt(dX*dX + dY*dY + dZ*dZ);  rij3 = rij*rij*rij;
            dVx[i][1] += (Planets[j]->Mass)*dX / rij3;
            dVy[i][1] += (Planets[j]->Mass)*dY / rij3;
            dVz[i][1] += (Planets[j]->Mass)*dZ / rij3;
        }
        dVx[i][1] *= dt*Gamma;  dVy[i][1] *= dt*Gamma;  dVz[i][1] *= dt*Gamma;
    }
//	 2 :
	for (i=0; i<Count; i++) {
    	dx[i][2] = dt*(dVx[i][0] + dVx[i][1]);
        dy[i][2] = dt*(dVy[i][0] + dVy[i][1]);
        dz[i][2] = dt*(dVz[i][0] + dVz[i][1]);

        dVx[i][2] = 0;  dVy[i][2] = 0;  dVz[i][2] = 0;
        for (j=0;j<Count;j++) if (i!=j) {
           	dX = dx[j][0]-dx[i][0] + (dx[j][1]-dx[i][1]);
            dY = dy[j][0]-dy[i][0] + (dy[j][1]-dy[i][1]);
            dZ = dz[j][0]-dz[i][0] + (dz[j][1]-dz[i][1]);
          	rij = sqrt(dX*dX + dY*dY + dZ*dZ);   rij3 = rij*rij*rij;
            dVx[i][2] += (Planets[j]->Mass)*dX / rij3;
            dVy[i][2] += (Planets[j]->Mass)*dY / rij3;
            dVz[i][2] += (Planets[j]->Mass)*dZ / rij3;
        }
        dVx[i][2] *= dt*Gamma;  dVy[i][2] *= dt*Gamma;  dVz[i][2] *= dt*Gamma;
    }
//	     :
	for (i=0; i<Count; i++) {
    	Planets[i]->NextX  = dx[i][0] + (dx[i][1] + dx[i][2])/2;
        Planets[i]->NextY  = dy[i][0] + (dy[i][1] + dy[i][2])/2;
        Planets[i]->NextZ  = dz[i][0] + (dz[i][1] + dz[i][2])/2;
        Planets[i]->NextVx = dVx[i][0] + (dVx[i][1] + dVx[i][2])/2;
        Planets[i]->NextVy = dVy[i][0] + (dVy[i][1] + dVy[i][2])/2;
        Planets[i]->NextVz = dVz[i][0] + (dVz[i][1] + dVz[i][2])/2;
    }
//	     Next_:
    Tau = dt;
//	  :
	if (DetectImpacts) {
		for (i=0; i<Count; i++)
        	for (j=i+1; j<Count; j++)  NoImpacts[i][j] = false;
    }
};

// ----------------------------------------------------------------------------
//      3(3)   dt.   
//  Next_ .

void TUniverse::RungeKutta3Step(double dt)
{
	int i,j,k;
    double s, rij, rij3, dX,dY,dZ;
    s = 2/3.0;

//	      0- :
	for (i=0; i<Count; i++) {
    	dx[i][0]  = Planets[i]->X;		dVx[i][0] = Planets[i]->Vx;
        dy[i][0]  = Planets[i]->Y;		dVy[i][0] = Planets[i]->Vy;
        dz[i][0]  = Planets[i]->Z;		dVz[i][0] = Planets[i]->Vz;
    }
//	 1- :
	for (i=0; i<Count; i++) {
    	dx[i][1]  = dt*dVx[i][0];		dVx[i][1] = 0;
        dy[i][1]  = dt*dVy[i][0];		dVy[i][1] = 0;
        dz[i][1]  = dt*dVz[i][0];		dVz[i][1] = 0;

        for (j=0;j<Count;j++) if (i!=j) {
        	dX = dx[j][0]-dx[i][0];
            dY = dy[j][0]-dy[i][0];
            dZ = dz[j][0]-dz[i][0];
        	rij = sqrt(dX*dX + dY*dY + dZ*dZ);  rij3 = rij*rij*rij;
            dVx[i][1] += (Planets[j]->Mass)*dX / rij3;
            dVy[i][1] += (Planets[j]->Mass)*dY / rij3;
            dVz[i][1] += (Planets[j]->Mass)*dZ / rij3;
        }
        dVx[i][1] *= dt*Gamma;  dVy[i][1] *= dt*Gamma;  dVz[i][1] *= dt*Gamma;
    }
//	 2,3 :
    for (k=2; k<=3; k++) {
		for (i=0; i<Count; i++) {
	    	dx[i][k] = dt*(dVx[i][0] + s*dVx[i][k-1]);
	        dy[i][k] = dt*(dVy[i][0] + s*dVy[i][k-1]);
	        dz[i][k] = dt*(dVz[i][0] + s*dVz[i][k-1]);

	        dVx[i][k] = 0;  dVy[i][k] = 0;  dVz[i][k] = 0;
	        for (j=0;j<Count;j++) if (i!=j) {
            	dX = dx[j][0]-dx[i][0]+s*(dx[j][k-1]-dx[i][k-1]);
                dY = dy[j][0]-dy[i][0]+s*(dy[j][k-1]-dy[i][k-1]);
                dZ = dz[j][0]-dz[i][0]+s*(dz[j][k-1]-dz[i][k-1]);
            	rij = sqrt(dX*dX + dY*dY + dZ*dZ);  rij3 = rij*rij*rij;
                dVx[i][k] += (Planets[j]->Mass)*dX / rij3;
                dVy[i][k] += (Planets[j]->Mass)*dY / rij3;
                dVz[i][k] += (Planets[j]->Mass)*dZ / rij3;
            }
	        dVx[i][k] *= dt*Gamma;  dVy[i][k] *= dt*Gamma;  dVz[i][k] *= dt*Gamma;
	    }
    }
//	     :
	for (i=0; i<Count; i++) {
    	Planets[i]->NextX  = dx[i][0] + (2*dx[i][1] + 3*dx[i][2] + 3*dx[i][3])/8;
        Planets[i]->NextY  = dy[i][0] + (2*dy[i][1] + 3*dy[i][2] + 3*dy[i][3])/8;
        Planets[i]->NextZ  = dz[i][0] + (2*dz[i][1] + 3*dz[i][2] + 3*dz[i][3])/8;
        Planets[i]->NextVx = dVx[i][0] + (2*dVx[i][1] + 3*dVx[i][2] + 3*dVx[i][3])/8;
        Planets[i]->NextVy = dVy[i][0] + (2*dVy[i][1] + 3*dVy[i][2] + 3*dVy[i][3])/8;
        Planets[i]->NextVz = dVz[i][0] + (2*dVz[i][1] + 3*dVz[i][2] + 3*dVz[i][3])/8;
    }
//	     Next_:
    Tau = dt;
//	  :
	if (DetectImpacts) {
		for (i=0; i<Count; i++)
        	for (j=i+1; j<Count; j++)  NoImpacts[i][j] = false;
    }
};

// ----------------------------------------------------------------------------
//    :

void TUniverse::Impact(int i0, int j0)
{
	if (!(0<=i0 && i0<Count && 0<=j0 && j0<Count)) return;
    TPlanet *P1 = Planets[i0];
    TPlanet *P2 = Planets[j0];
//	  :
	double r = sqrt((P2->X-P1->X)*(P2->X-P1->X) + (P2->Y-P1->Y)*(P2->Y-P1->Y) + (P2->Z-P1->Z)*(P2->Z-P1->Z));
    if (r<MINDOUBLE) return;
//	    P1  P2:
	double cx = (P2->X - P1->X)/r;
	double cy = (P2->Y - P1->Y)/r;
	double cz = (P2->Z - P1->Z)/r;
//	    (     Z):
    double d  = sqrt(cy*cy + cz*cz);
	double a11,a12,a13,a21,a22,a23,a31,a32,a33;
    if (d > MINDOUBLE) {
    	a11 = d;   a12 = -1/d*(cx*cy);  a13 = -1/d*(cx*cz);
        a21 = 0;   a22 = 1/d*(cz);      a23 = -1/d*(cy);
        a31 = cx;  a32 = cy;            a33 = cz;
    }
    else {
    	a11 = 0;  a12 = 0;  a13 = -1;
        a21 = 0;  a22 = 1;  a23 =  0;
        a31 = 1;  a32 = 0;  a33 =  0;
    }
//	       :
	double U1x = (P1->Vx)*a11 + (P1->Vy)*a12 + (P1->Vz)*a13;
    double U1y = (P1->Vx)*a21 + (P1->Vy)*a22 + (P1->Vz)*a23;
    double U1z = (P1->Vx)*a31 + (P1->Vy)*a32 + (P1->Vz)*a33;
	double U2x = (P2->Vx)*a11 + (P2->Vy)*a12 + (P2->Vz)*a13;
    double U2y = (P2->Vx)*a21 + (P2->Vy)*a22 + (P2->Vz)*a23;
    double U2z = (P2->Vx)*a31 + (P2->Vy)*a32 + (P2->Vz)*a33;
//	      ..    Z,  
//	  X, Y  .       
//	     Z. ,    .. 2- 
//	 1-,     :
    if (U2z>=0 && U1z-U2z<=0) { NoImpacts[i0][j0] = true;  return; }
    if (U1z<=0 && U2z-U1z>=0) { NoImpacts[i0][j0] = true;  return; }
//	  ( ):
	double m1 = P1->Mass;
    double m2 = P2->Mass;
//	        Next_ ()
//	(   ):
	double I1x = m1*(P1->Vx) + m2*(P2->Vx);
	double I1y = m1*(P1->Vy) + m2*(P2->Vy);
	double I1z = m1*(P1->Vz) + m2*(P2->Vz);
	double I2x = m1*(P1->NextVx) + m2*(P2->NextVx);
	double I2y = m1*(P1->NextVy) + m2*(P2->NextVy);
	double I2z = m1*(P1->NextVz) + m2*(P2->NextVz);
/*
//	          Next_ ()
//  (   ):
	double Xc1 = (m1*(P1->X) + m2*(P2->X))/(m1+m2);
	double Yc1 = (m1*(P1->Y) + m2*(P2->Y))/(m1+m2);
	double Zc1 = (m1*(P1->Z) + m2*(P2->Z))/(m1+m2);
	double Xc2 = (m1*(P1->NextX) + m2*(P2->NextX))/(m1+m2);
	double Yc2 = (m1*(P1->NextY) + m2*(P2->NextY))/(m1+m2);
	double Zc2 = (m1*(P1->NextZ) + m2*(P2->NextZ))/(m1+m2);
//	          Next_ ()
//	(   ):
	double K1x = m1*((P1->Y - Yc1)*(P1->Vz) - (P1->Z - Zc1)*(P1->Vy)) +
            	 m2*((P2->Y - Yc1)*(P2->Vz) - (P2->Z - Zc1)*(P2->Vy));
	double K1y = m1*((P1->Z - Zc1)*(P1->Vx) - (P1->X - Xc1)*(P1->Vz)) +
            	 m2*((P2->Z - Zc1)*(P2->Vx) - (P2->X - Xc1)*(P2->Vz));
	double K1z = m1*((P1->X - Xc1)*(P1->Vy) - (P1->Y - Yc1)*(P1->Vx)) +
            	 m2*((P2->X - Xc1)*(P2->Vy) - (P2->Y - Yc1)*(P2->Vx));
	double K2x = m1*((P1->NextY - Yc2)*(P1->NextVz) - (P1->NextZ - Zc2)*(P1->NextVy)) +
            	 m2*((P2->NextY - Yc2)*(P2->NextVz) - (P2->NextZ - Zc2)*(P2->NextVy));
	double K2y = m1*((P1->NextZ - Zc2)*(P1->NextVx) - (P1->NextX - Xc2)*(P1->NextVz)) +
            	 m2*((P2->NextZ - Zc2)*(P2->NextVx) - (P2->NextX - Xc2)*(P2->NextVz));
	double K2z = m1*((P1->NextX - Xc2)*(P1->NextVy) - (P1->NextY - Yc2)*(P1->NextVx)) +
            	 m2*((P2->NextX - Xc2)*(P2->NextVy) - (P2->NextY - Yc2)*(P2->NextVx));
//	       
//	(   ):
	double Jx = m1*((P1->Y - Yc1)*(P1->Y - Yc1) + (P1->Z - Zc1)*(P1->Z - Zc1)) +
                m2*((P2->Y - Yc1)*(P2->Y - Yc1) + (P2->Z - Zc1)*(P2->Z - Zc1));
	double Jy = m1*((P1->X - Xc1)*(P1->X - Xc1) + (P1->Z - Zc1)*(P1->Z - Zc1)) +
                m2*((P2->X - Xc1)*(P2->X - Xc1) + (P2->Z - Zc1)*(P2->Z - Zc1));
	double Jz = m1*((P1->X - Xc1)*(P1->X - Xc1) + (P1->Y - Yc1)*(P1->Y - Yc1)) +
                m2*((P2->X - Xc1)*(P2->X - Xc1) + (P2->Y - Yc1)*(P2->Y - Yc1));
*/
//	      Z  .  
//	  -   .  ,   -
//	   .    ,   Alpha:
//		P = m1*U1z + m2*U2z;
//		Q = (P*P + Alpha*m1*m2*(U1z-U2z)*(U1z-U2z)) / (2*(m1+m2));
//	      :
//		m1*V1z + m2*V2z = P
//		0.5*(m1*V1z^2 + m2*V2z^2) = Q
//	   Z   (  ..):
	double A = sqrt((Alpha>0) ? Alpha : -Alpha);
	double V1z = (U1z*(m1-A*m2) + U2z*m2*(1+A)) / (m1+m2);
    double V2z = (U2z*(m2-A*m1) + U1z*m1*(1+A)) / (m1+m2);
//	        ..:
	P1->Vx = U1x*a11 + U1y*a21 + V1z*a31;  P2->Vx = U2x*a11 + U2y*a21 + V2z*a31;
    P1->Vy = U1x*a12 + U1y*a22 + V1z*a32;  P2->Vy = U2x*a12 + U2y*a22 + V2z*a32;
    P1->Vz = U1x*a13 + U1y*a23 + V1z*a33;  P2->Vz = U2x*a13 + U2y*a23 + V2z*a33;
//	     :
//	( ,       Tau, .. 
//	Next_      .   
//	 ,  , .    ,   
//	   .    
//	  , ,    Next_    
//	Next_.    .)
//	 :
	double DeltaVx = (I2x - I1x)/(m1+m2);	//	double Delta_x = (K2x-K1x)/Jx;
	double DeltaVy = (I2y - I1y)/(m1+m2);	//	double Delta_y = (K2y-K1y)/Jy;
	double DeltaVz = (I2z - I1z)/(m1+m2);	//	double Delta_z = (K2z-K1z)/Jz;
	P1->Vx += DeltaVx;	// + Delta_y*(P1->Z - Zc1) - Delta_z*(P1->Y - Yc1);
	P1->Vy += DeltaVy;	// - Delta_x*(P1->Z - Zc1) + Delta_z*(P1->X - Xc1);
    P1->Vz += DeltaVz;	// + Delta_x*(P1->Y - Yc1) - Delta_y*(P1->X - Xc1);
	P2->Vx += DeltaVx;	// + Delta_y*(P2->Z - Zc1) - Delta_z*(P2->Y - Yc1);
	P2->Vy += DeltaVy;	// - Delta_x*(P2->Z - Zc1) + Delta_z*(P2->X - Xc1);
    P2->Vz += DeltaVz;	// + Delta_x*(P2->Y - Yc1) - Delta_y*(P2->X - Xc1);
//	,       :
	P1->NextVx = P1->Vx;                 P2->NextVx = P2->Vx;
	P1->NextVy = P1->Vy;                 P2->NextVy = P2->Vy;
	P1->NextVz = P1->Vz;                 P2->NextVz = P2->Vz;
   	P1->NextX = P1->X + Tau*(P1->Vx);    P2->NextX = P2->X + Tau*(P2->Vx);
   	P1->NextY = P1->Y + Tau*(P1->Vy);    P2->NextY = P2->Y + Tau*(P2->Vy);
   	P1->NextZ = P1->Z + Tau*(P1->Vz);    P2->NextZ = P2->Z + Tau*(P2->Vz);
//	         ,     .
//	   ,  :
   	int i,j;
   	for (i=0; i<i0; i++) NoImpacts[i][i0] = false;
   	for (i=0; i<j0; i++) NoImpacts[i][j0] = false;
   	for (j=i0+1; j<Count; j++) NoImpacts[i0][j] = false;
   	for (j=j0+1; j<Count; j++) NoImpacts[j0][j] = false;
//	        ,      - :
	NoImpacts[i0][j0] = true;
};

// ----------------------------------------------------------------------------
//         
// .    .  , 
// .    ,  Tau -    .

int TUniverse::MoveTo1stImpact()
{
	if (Tau == 0) return 0;

    double dx0,dy0,dz0;	//     
    double dx1,dy1,dz1;	//     
    double R,r_min;		//  
    double L;			//    
	double a,b2,D4;		// , /4   
    double s,_s,s1,s2;	//  [0;1]  
    double eps = 0.01;	//       
    int    i0,j0;		//    
    int    Drown = 0;	//  
    int i,j;

//	   :
	s = 1;  i0 = -1;  j0 = -1;
    for (i=0; i<Count; i++)
    	for (j=i+1; j<Count; j++) {
        	if (NoImpacts[i][j]) continue;
        	dx0 = Planets[i]->X - Planets[j]->X;
            dy0 = Planets[i]->Y - Planets[j]->Y;
            dz0 = Planets[i]->Z - Planets[j]->Z;
            dx1 = Planets[i]->NextX - Planets[j]->NextX;
            dy1 = Planets[i]->NextY - Planets[j]->NextY;
            dz1 = Planets[i]->NextZ - Planets[j]->NextZ;
            R = Planets[i]->Radius + Planets[j]->Radius;
        	a  = (dx1-dx0)*(dx1-dx0) + (dy1-dy0)*(dy1-dy0) + (dz1-dz0)*(dz1-dz0);
            D4 = a*R*R - (dx0*dy1 - dx1*dy0)*(dx0*dy1 - dx1*dy0) -
                         (dx0*dz1 - dx1*dz0)*(dx0*dz1 - dx1*dz0) -
                         (dy0*dz1 - dy1*dz0)*(dy0*dz1 - dy1*dz0);
            //	   (D4 ~= 0  ):
            if (D4>0 && a>MINDOUBLE) {
            	b2 = dx0*(dx1-dx0) + dy0*(dy1-dy0) + dz0*(dz1-dz0);
            	//	    ( - ):
                s1 = (-b2+sqrt(D4))/a;    s2 = (-b2-sqrt(D4))/a;
            	_s = (s1 <= s2) ? s1 : s2;
                //	    [0;1]   ,  :
                if (0<_s && _s<=1) {
                	if (_s<s) { s = _s;  i0 = i;  j0 = j; }
                    continue;
                }
                if (_s>1) { NoImpacts[i][j] = true; continue; }
                if (_s<0) {
                	L = sqrt(dx0*dx0 + dy0*dy0 + dz0*dz0);
                    r_min = (Planets[i]->Radius <= Planets[j]->Radius) ? Planets[i]->Radius : Planets[j]->Radius;
                	//     -  . :
                	if (R - L > eps*r_min) {
                    	if (L>MINDOUBLE) {
						//	:
                        	double dLx,dLy,dLz;
                        	double mi = Planets[i]->Mass;
                        	double mj = Planets[j]->Mass;
                        	dLx = (R-L)*dx0/L;  dLy = (R-L)*dy0/L;  dLz = (R-L)*dz0/L;
                        	Planets[i]->X += dLx*mj/(mi + mj);  Planets[j]->X -= dLx*mi/(mi + mj);
                        	Planets[i]->Y += dLy*mj/(mi + mj);  Planets[j]->Y -= dLy*mi/(mi + mj);
                        	Planets[i]->Z += dLz*mj/(mi + mj);  Planets[j]->Z -= dLz*mi/(mi + mj);
						//	:
                            Impact(i,j);
                            Drown = 1;
                        }
                	}
                    else NoImpacts[i][j] = true;
                }
            }
            else NoImpacts[i][j] = true;
        }

//	     
//	(       ):
	for (i=0; i<Count; i++) {
    	Planets[i]->X = (1-s)*Planets[i]->X + s*(Planets[i]->NextX);
        Planets[i]->Y = (1-s)*Planets[i]->Y + s*(Planets[i]->NextY);
        Planets[i]->Z = (1-s)*Planets[i]->Z + s*(Planets[i]->NextZ);
        Planets[i]->Vx = (1-s)*(Planets[i]->Vx) + s*(Planets[i]->NextVx);
        Planets[i]->Vy = (1-s)*(Planets[i]->Vy) + s*(Planets[i]->NextVy);
        Planets[i]->Vz = (1-s)*(Planets[i]->Vz) + s*(Planets[i]->NextVz);
    }
    Tau = Tau*(1-s);
//	  :
	if (i0>-1 && j0>-1) {
    	Impact(i0,j0);
    	return 1;
    }
//	  :
	return Drown;
};

// ----------------------------------------------------------------------------
//  :      .

void TUniverse::NextStep(double dt)
{
	int Moved, i;
    int Sel = -1, NSel = 0;
    //  :
    double Xc,Yc,Zc, M,m;
    //        :
    double X1,Y1,Z1;

//	        :
	if (Camera->RotateWith && Active > -1) {
	//	   :
		Xc=0, Yc=0, Zc=0; M = 0;
	    for (i=0; i<Count; i++)
			if (Planets[i]->Selected) {
	        	m = Planets[i]->Mass;
	            Xc += (Planets[i]->X)*m;
	            Yc += (Planets[i]->Y)*m;
	            Zc += (Planets[i]->Z)*m;
	        	M += m;
                Sel = i;  NSel++;
	        }
	    if (M > MINDOUBLE) {
	    	Xc /= M;  Yc /= M;  Zc /= M;
    		X1 = Planets[Active]->X - Xc;
    		Y1 = Planets[Active]->Y - Yc;
    		Z1 = Planets[Active]->Z - Zc;
	    }
    }

//	  :
	switch (Method) {
    	//	 - 2(2):
    	case 0 : RungeKutta2Step(dt); break;
        //	 - 3(3):
        case 1 : RungeKutta3Step(dt); break;
        //	 - 4(4) "":
		case 2 : RungeKutta4Step(dt); break;
        //	 - 5(6):
        case 3 : XRungeKuttaStep(c5, b5, a5, 6, dt); break;
        //	  6(7):
        case 4 : XRungeKuttaStep(c6, b6, a6, 7, dt); break;
        //	  7(11):
        case 5 : XRungeKuttaStep(c7, b7, a7, 11, dt); break;
        //	 - 8(12):
        case 6 : XRungeKuttaStep(c8, b8, a8, 12, dt); break;
        //	 - 4(4) "":
        default : RungeKutta4Step(dt); break;
    }

//	       :
	if (DetectImpacts) {
    	do  Moved = MoveTo1stImpact();  while (Moved);
    }
//	  :
    else {
    	for (i=0; i<Count; i++) {
    		Planets[i]->X  = Planets[i]->NextX;
        	Planets[i]->Y  = Planets[i]->NextY;
        	Planets[i]->Z  = Planets[i]->NextZ;
        	Planets[i]->Vx = Planets[i]->NextVx;
        	Planets[i]->Vy = Planets[i]->NextVy;
        	Planets[i]->Vz = Planets[i]->NextVz;
    	}
    }

//	  :
	double dL2;  TPlanet *P;  T3DPoint *T;
	for (i=0; i<Count; i++)
    	if (Planets[i]->HasTrajectory) {
        	P = Planets[i];
            T = P->Trajectory->Last;
            if (T != NULL) {
            	dL2 = (T->X - P->X)*(T->X - P->X) + (T->Y - P->Y)*(T->Y - P->Y) + (T->Z - P->Z)*(T->Z - P->Z);
                if (dL2 > (P->Trajectory->dL)*(P->Trajectory->dL))
                	P->Trajectory->Put(P->X, P->Y, P->Z);
            }
            else P->Trajectory->Put(P->X, P->Y, P->Z);
        }

//	  :
	Time->Increment(dt);

//	     .
	//	   :
	Xc=0, Yc=0, Zc=0; M = 0;
    for (i=0; i<Count; i++)
		if (Planets[i]->Selected) {
        	m = Planets[i]->Mass;
            Xc += (Planets[i]->X)*m;
            Yc += (Planets[i]->Y)*m;
            Zc += (Planets[i]->Z)*m;
        	M += m;
        }
	//	:
    if (M > MINDOUBLE) {
    	Xc /= M;  Yc /= M;  Zc /= M;
        Camera->X0 = Xc + Camera->dX0;
        Camera->Y0 = Yc + Camera->dY0;
        Camera->Z0 = Zc + Camera->dZ0;
    }
	//	    :
	if (Camera->RotateWith && Active > -1 && M > MINDOUBLE) {
    	//  -      :
        if (Sel == Active && NSel <= 1) return;
    	//        :
    	double X2 = Planets[Active]->X - Xc;
    	double Y2 = Planets[Active]->Y - Yc;
    	double Z2 = Planets[Active]->Z - Zc;
        //  :
        double Wx = Y1*Z2 - Y2*Z1;
        double Wy = X2*Z1 - X1*Z2;
        double Wz = X1*Y2 - X2*Y1;
        //  :
        double Theta = acos((X1*X2 + Y1*Y2 + Z1*Z2)/sqrt((X1*X1 + Y1*Y1 + Z1*Z1)*(X2*X2 + Y2*Y2 + Z2*Z2)));
        Camera->RotateAround(Wx,Wy,Wz,Theta);
    }
};
